import itertools
from typing import Iterable

import numpy as np
from mvlearn.compose import SimpleSplitter
from sklearn import clone
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.model_selection import ParameterGrid
from sklearn.model_selection._search import (
    BaseSearchCV,
)
from sklearn.model_selection._search import ParameterSampler
from sklearn.pipeline import Pipeline
from sklearn.utils import check_random_state


def param2grid(params):
    params = {
        k: list(v) if (hasattr(v, "__iter__") and not isinstance(v, str)) else [v]
        for k, v in params.items()
    }
    for k, v in params.items():
        if any([hasattr(v_, "__iter__") and not isinstance(v_, str) for v_ in v]):
            params[k] = list(map(list, itertools.product(*v)))
    return ParameterGrid(params)


class ParameterSampler_(ParameterSampler):
    def __iter__(self):
        rng = check_random_state(self.random_state)
        for _ in range(self.n_iter):
            dist = rng.choice(self.param_distributions)
            # Always sort the keys of a dictionary, for reproducibility
            items = sorted(dist.items())
            params = dict()
            for k, v in items:
                # if v is iterable, then list comprehension else v
                if isinstance(v, Iterable) and not isinstance(v, str):
                    # use list comprehension to handle different types of values
                    params[k] = (
                        [self.return_param(v_) for v_ in v]
                        if isinstance(v, Iterable)
                        else self.return_param(v)
                    )
                else:
                    params[k] = self.return_param(v)
            yield params

    def return_param(self, v):
        rng = check_random_state(self.random_state)
        # use ternary operator to handle different types of values
        param = (
            v.rvs(random_state=rng)
            if hasattr(v, "rvs")
            else (
                v[rng.randint(len(v))]
                if isinstance(v, Iterable) and not isinstance(v, str)
                else v
            )
        )
        return param

    def __len__(self):
        """Number of points that will be sampled."""
        return self.n_iter


class BaseSearchCV(BaseSearchCV):
    def fit(self, views, y=None, *, groups=None, **fit_params):
        self.estimator = Pipeline(
            [
                ("splitter", SimpleSplitter([view.shape[1] for view in views])),
                ("estimator", clone(self.estimator)),
            ]
        )
        super().fit(np.hstack(views), y=y, groups=groups, **fit_params)
        self.estimator = self.estimator[1]
        self.best_estimator_ = self.best_estimator_[1]
        self.best_params_ = {
            key.split("estimator__")[1]: val for key, val in self.best_params_.items()
        }
        return self


class GridSearchCV(GridSearchCV, BaseSearchCV):
    def _run_search(self, evaluate_candidates):
        """Search all candidates in param_grid"""
        if not isinstance(self.param_grid, ParameterGrid):
            param_grid = param2grid(self.param_grid)
        else:
            param_grid = self.param_grid
        param_grid.param_grid = [
            {f"estimator__{key}": val for key, val in subgrid.items()}
            for subgrid in param_grid.param_grid
        ]
        evaluate_candidates(param_grid)


class RandomizedSearchCV(RandomizedSearchCV, BaseSearchCV):
    def _run_search(self, evaluate_candidates):
        self.param_distributions = {
            f"estimator__{key}": val for key, val in self.param_distributions.items()
        }
        """Search n_iter candidates from param_distributions"""
        evaluate_candidates(
            ParameterSampler_(
                self.param_distributions, self.n_iter, random_state=self.random_state
            )
        )



from ._dcca import DCCA
from ._dcca_barlow_twins import BarlowTwins
from ._dcca_ey import DCCA_EY
from ._dcca_gha import DCCA_GHA
from ._dcca_noi import DCCA_NOI
from ._dcca_sdl import DCCA_SDL
from ._dcca_svd import DCCA_SVD
from ._dgcca import DGCCA
from ._dtcca import DTCCA

__all__ = [
    "DCCA",
    "DCCA_EY",
    "DCCA_GHA",
    "DCCA_SVD",
    "DCCA_NOI",
    "DCCA_SDL",
    "DTCCA",
    "BarlowTwins",
    "DGCCA",
]



import itertools
from typing import List, Union

import numpy as np
from scipy.linalg import block_diag
from sklearn.datasets import make_spd_matrix
from sklearn.utils.validation import check_random_state

from cca_zoo.utils import _process_parameter


class LinearSimulatedData:
    """
    Class for generating simulated data for linear model with multiple views.
    """

    def __init__(
        self,
        view_features: List[int],
        latent_dims: int = 1,
        view_sparsity: Union[List[float], float] = None,
        correlation: Union[List[float], float] = 0.99,
        structure: str = "random",
        positive: Union[bool, List[bool]] = False,
        random_state: Union[int, np.random.RandomState] = None,
    ):
        self.view_features = view_features
        self.latent_dims = latent_dims
        self.random_state = check_random_state(random_state)
        self.correlation = _process_parameter(
            "correlation", correlation, 0.99, latent_dims
        )
        self.view_sparsity = _process_parameter(
            "view_sparsity", view_sparsity, 0.5, len(view_features)
        )
        self.structure = _process_parameter(
            "structure", structure, "random", len(view_features)
        )
        self.positive = _process_parameter(
            "positive", positive, False, len(view_features)
        )

        cov_matrices, self.true_features = self._generate_covariance_matrices()
        joint_cov = self._generate_joint_covariance(cov_matrices)
        self.chol = np.linalg.cholesky(joint_cov)

    def _generate_covariance_matrix(self, view_features, view_structure):
        """Generates a covariance matrix for a single view."""
        if view_structure == "identity":
            cov = np.eye(view_features)
        else:
            cov = make_spd_matrix(view_features)

        var = np.diag(cov)
        cov = cov / np.sqrt(var)
        cov = cov.T / np.sqrt(var)

        return cov

    def _generate_joint_covariance(self, cov_matrices):
        """Generates a joint covariance matrix for all views."""
        joint_cov = block_diag(*cov_matrices)
        split_points = np.concatenate(([0], np.cumsum(self.view_features)))

        for i, j in itertools.combinations(range(len(split_points) - 1), 2):
            cross_cov = self._compute_cross_cov(cov_matrices, i, j)
            joint_cov[
                split_points[i] : split_points[i + 1],
                split_points[j] : split_points[j + 1],
            ] = cross_cov
            joint_cov[
                split_points[j] : split_points[j + 1],
                split_points[i] : split_points[i + 1],
            ] = cross_cov.T

        return joint_cov

    def _compute_cross_cov(self, cov_matrices, i, j):
        """Computes the cross-covariance matrix for a pair of views."""
        cross_cov = np.zeros((self.view_features[i], self.view_features[j]))

        for _ in range(self.latent_dims):
            outer_product = np.outer(
                self.true_features[i][:, _], self.true_features[j][:, _]
            )
            cross_cov += (
                cov_matrices[i]
                @ (self.correlation[_] * outer_product)
                @ cov_matrices[j]
            )

        return cross_cov

    def _generate_covariance_matrices(self):
        """Generates a list of covariance matrices and true features for each view."""
        cov_matrices = [
            self._generate_covariance_matrix(f, s)
            for f, s in zip(self.view_features, self.structure)
        ]
        true_features = [
            self._generate_true_feature(cov, s, pos)
            for cov, s, pos in zip(cov_matrices, self.view_sparsity, self.positive)
        ]
        return cov_matrices, true_features

    def _generate_true_feature(self, cov, sparsity, is_positive):
        """Generates a true feature matrix for a single view."""
        weights = self._generate_weights(cov.shape[0])
        weights = self._apply_sparsity(weights, sparsity)

        if is_positive:
            weights = np.abs(weights)

        weights = self._decorrelate_weights(weights, cov)
        return weights / np.sqrt(np.diag(weights.T @ cov @ weights))

    def _generate_weights(self, view_features):
        return self.random_state.randn(view_features, self.latent_dims)

    def _apply_sparsity(self, weights, sparsity):
        if sparsity <= 1:
            sparsity = np.ceil(sparsity * weights.shape[0]).astype(int)

        mask = self._generate_sparsity_mask(weights.shape, sparsity)
        return weights * mask

    @staticmethod
    def _generate_sparsity_mask(shape, sparsity):
        mask_elements = [0] * (shape[0] - sparsity) + [1] * sparsity
        mask = np.stack([mask_elements] * shape[1]).T
        np.random.shuffle(mask)
        return mask.astype(bool)

    @staticmethod
    def _decorrelate_weights(weights, cov):
        product = weights.T @ cov @ weights
        for k in range(1, product.shape[0]):
            weights[:, k:] -= np.outer(
                weights[:, k - 1], product[k - 1, k:] / product[k - 1, k - 1]
            )
            product = weights.T @ cov @ weights
        return weights

    def sample(self, n_samples: int):
        random_data = self.random_state.randn(self.chol.shape[0], n_samples)
        samples = (self.chol @ random_data).T
        return np.split(samples, np.cumsum(self.view_features)[:-1], axis=1)



# Import the necessary modules
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.ticker as mtick

from cca_zoo.utils.check_values import check_seaborn_support


class ExplainedCovarianceDisplay:
    """
    Display the explained covariance of the latent variables of the views.

    Parameters
    ----------
    explained_covariance_train : np.ndarray
        The explained covariance of the train data.
    explained_covariance_test : np.ndarray
        The explained covariance of the test data.
    ratio : bool
        Whether to plot the ratio of explained covariance or not.
    **kwargs : dict
        Keyword arguments to be passed to the seaborn lineplot.

    Attributes
    ----------
    figure_ : matplotlib.pyplot.figure
        The figure of the plot.

    Examples
    --------
    >>> from cca_zoo.visualisation import ExplainedCovarianceDisplay
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from cca_zoo.linear import MCCA
    >>>
    >>> # Generate Sample Data
    >>> # --------------------
    >>> X = np.random.rand(100, 10)
    >>> Y = np.random.rand(100, 10)
    >>>
    >>> # Splitting the data into training and testing sets
    >>> X_train, X_test = X[:50], X[50:]
    >>> Y_train, Y_test = Y[:50], Y[50:]
    >>>
    >>> views = [X_train, Y_train]
    >>> test_views = [X_test, Y_test]
    >>>
    >>> # Train an MCCA Model
    >>> # -------------------
    >>> mcca = MCCA(latent_dimensions=2)
    >>> mcca.fit(views)
    >>>
    >>> # %%
    >>> # Plotting the Explained Covariance
    >>> # ---------------------------------
    >>> ExplainedCovarianceDisplay.from_estimator(mcca, views, test_views=test_views).plot()
    >>> plt.show()

    """

    def __init__(
        self,
        explained_covariance_train,
        explained_covariance_test=None,
        ratio=True,
        **kwargs
    ):
        self.explained_covariance_train = explained_covariance_train
        self.explained_covariance_test = explained_covariance_test
        self.ratio = ratio
        self.kwargs = kwargs

    def _validate_plot_params(self):
        check_seaborn_support("CorrelationHeatmapDisplay")

    @classmethod
    def from_estimator(cls, model, train_views, test_views=None, ratio=True, **kwargs):
        # explained_covariance_train will be a numpy array of shape (latent_dimensions,len(train_views))
        if ratio:
            explained_covariance_train = model.explained_covariance_ratio(train_views)
        else:
            explained_covariance_train = model.explained_covariance(train_views)
        if test_views is not None:
            if ratio:
                explained_covariance_test = model.explained_covariance_ratio(test_views)
            else:
                explained_covariance_test = model.explained_covariance(test_views)
        else:
            explained_covariance_test = None
        if ratio:
            return cls.from_explained_covariance_ratio(
                explained_covariance_train, explained_covariance_test, **kwargs
            )
        else:
            return cls.from_explained_covariance(
                explained_covariance_train, explained_covariance_test, **kwargs
            )

    @classmethod
    def from_explained_covariance(
        cls, explained_covariance_train, explained_covariance_test=None, **kwargs
    ):
        return cls(
            explained_covariance_train, explained_covariance_test, ratio=False, **kwargs
        )

    @classmethod
    def from_explained_covariance_ratio(
        cls, explained_covariance_train, explained_covariance_test=None, **kwargs
    ):
        return cls(
            explained_covariance_train, explained_covariance_test, ratio=True, **kwargs
        )

    def plot(self, ax=None):
        self._validate_plot_params()
        # Use seaborn lineplot with hue='Train' to plot the train and test data
        data = pd.DataFrame(self.explained_covariance_train, columns=["value"])
        data["Mode"] = "Train"  # Add a column indicating train data
        data.index.name = "Latent dimension"
        if self.explained_covariance_test is not None:
            data_test = pd.DataFrame(self.explained_covariance_test, columns=["value"])
            data_test["Mode"] = "Test"  # Add a column indicating test data
            data_test.index.name = "Latent dimension"
            data = pd.concat([data, data_test])  # Concatenate the two dataframes
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 5))
        else:
            fig = ax.get_figure()
        sns.lineplot(
            data=data, x="Latent dimension", y="value", style="Mode", marker="o", ax=ax
        )
        ax.set_xlabel("Latent dimension")
        if self.ratio:
            ax.set_ylabel("Explained covariance %")
            ax.yaxis.set_major_formatter(mtick.PercentFormatter(1.0))
        else:
            ax.set_ylabel("Explained covariance")
        ax.set_title("Explained covariance")
        # Set x-ticks to integers
        ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True))
        plt.tight_layout()
        self.figure_ = fig
        return self



from typing import Iterable

import numpy as np
import tensorly as tl
from scipy.linalg import sqrtm
from tensorly.decomposition import parafac

from cca_zoo.linear._mcca import MCCA


class TCCA(MCCA):
    r"""
    A class used to fit TCCA model. This model extends MCCA to higher order correlations by using tensor products of the views.

    The objective function of TCCA is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^T\otimes w_2^TX_2^T\otimes \cdots \otimes w_m^TX_m^Tw  \}\\

        \text{subject to:}

        w_i^TX_i^TX_iw_i=1

    where :math:`\otimes` denotes the Kronecker product.

    References
    ----------
    Kim, Tae-Kyun, Shu-Fai Wong, and Roberto Cipolla. "Tensor canonical correlation analysis for action classification." 2007 IEEE Conference on Computer Vision and Pattern Recognition. IEEE, 2007

    Examples
    --------
    >>> from cca_zoo.linear import TCCA
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> X3 = rng.random((10,5))
    >>> model = TCCA()
    >>> model.fit((X1,X2,X3)).score((X1,X2,X3))
    """

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        views = self._validate_data(views)
        self._check_params()
        # returns whitened views along with whitening matrices
        whitened_views, covs_invsqrt = self._setup_tensor(views)
        # The idea here is to form a matrix with M dimensions one for each view where at index
        # M[p_i,p_j,p_k...] we have the sum over n samples of the product of the pth feature of the
        # ith, jth, kth view etc.
        for i, el in enumerate(whitened_views):
            # To achieve this we start with the first view so M is nxp.
            if i == 0:
                M = el
            # For the remaining views we expand their dimensions to match M i.e. nx1x...x1xp
            else:
                for _ in range(len(M.shape) - 1):
                    el = np.expand_dims(el, 1)
                # Then we perform an outer product by expanding the dimensionality of M and
                # outer product with the expanded el
                M = np.expand_dims(M, -1) @ el
        M = np.mean(M, 0)
        tl.set_backend("numpy")
        M_parafac = parafac(M, self.latent_dimensions, verbose=False)
        self.weights = [
            cov_invsqrt @ fac
            for i, (view, cov_invsqrt, fac) in enumerate(
                zip(whitened_views, covs_invsqrt, M_parafac.factors)
            )
        ]
        return self

    def correlations(self, views: Iterable[np.ndarray], **kwargs):
        """
        Predicts the correlation for the given data using the fit model

        :param views: list/tuple of numpy arrays or array likes with the same number of rows (samples)
        :param kwargs: any additional keyword arguments required by the given model
        """
        transformed_views = self.transform(views, **kwargs)
        transformed_views = [
            transformed_view - transformed_view.mean(axis=0)
            for transformed_view in transformed_views
        ]
        multiplied_views = np.stack(transformed_views, axis=0).prod(axis=0).sum(axis=0)
        norms = np.stack(
            [
                np.linalg.norm(transformed_view, axis=0)
                for transformed_view in transformed_views
            ],
            axis=0,
        ).prod(axis=0)
        corrs = multiplied_views / norms
        return corrs

    def score(self, views: Iterable[np.ndarray], **kwargs):
        """
        Returns the higher order correlations in each dimension

        :param views: list/tuple of numpy arrays or array likes with the same number of rows (samples)
        :param kwargs: any additional keyword arguments required by the given model
        """
        dim_corrs = self.correlations(views, **kwargs)
        return dim_corrs

    def _setup_tensor(self, views: Iterable[np.ndarray], **kwargs):
        covs = [
            (1 - self.c[i]) * np.cov(view, rowvar=False)
            + self.c[i] * np.eye(view.shape[1])
            for i, view in enumerate(views)
        ]
        smallest_eigs = [
            min(0, np.linalg.eigvalsh(cov).min()) - self.eps for cov in covs
        ]
        covs = [
            cov - smallest_eig * np.eye(cov.shape[0])
            for cov, smallest_eig in zip(covs, smallest_eigs)
        ]
        covs_invsqrt = [np.linalg.inv(sqrtm(cov).real) for cov in covs]
        views = [
            train_view @ cov_invsqrt
            for train_view, cov_invsqrt in zip(views, covs_invsqrt)
        ]
        return views, covs_invsqrt

    def _more_tags(self):
        return {"multiview": True}



from abc import abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pytorch_lightning as pl
import torch
from torch.optim.lr_scheduler import CosineAnnealingLR, MultiStepLR

from cca_zoo._base import BaseModel


class BaseDeep(pl.LightningModule, BaseModel):
    """A base class for deep learning linear using PyTorch Lightning."""

    def __init__(
        self,
        latent_dimensions: int,
        optimizer: str = "adam",
        scheduler: Optional[str] = None,
        lr: float = 1e-2,
        weight_decay: float = 0,
        extra_optimizer_kwargs: Optional[Dict[str, Any]] = None,
        max_epochs: int = 1000,
        min_lr: float = 1e-9,
        lr_decay_steps: Optional[List[int]] = None,
        correlation: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__()
        if extra_optimizer_kwargs is None:
            extra_optimizer_kwargs = {}
        self.latent_dimensions = latent_dimensions
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.lr = lr
        self.weight_decay = weight_decay
        self.extra_optimizer_kwargs = extra_optimizer_kwargs
        self.max_epochs = max_epochs
        self.min_lr = min_lr
        self.lr_decay_steps = lr_decay_steps
        self.correlation = correlation

    @abstractmethod
    def forward(self, views: List[torch.Tensor], *args, **kwargs) -> List[torch.Tensor]:
        """Returns the latent representations for each view."""
        raise NotImplementedError

    @abstractmethod
    def loss(
        self, views: List[torch.Tensor], *args, **kwargs
    ) -> Dict[str, torch.Tensor]:
        """Returns the loss components for each view."""
        raise NotImplementedError

    def training_step(self, batch: Dict[str, Any], batch_idx: int) -> torch.Tensor:
        """Performs one step of training on a batch of views."""
        loss = self.loss(batch)
        for k, v in loss.items():
            # Use f-string instead of concatenation
            self.log(
                f"train/{k}",
                v,
                on_step=False,
                on_epoch=True,
                batch_size=batch["views"][0].shape[0],
            )
        return loss["objective"]

    def validation_step(self, batch: Dict[str, Any], batch_idx: int) -> torch.Tensor:
        """Performs one step of validation on a batch of views."""
        loss = self.loss(batch)
        for k, v in loss.items():
            # Use f-string instead of concatenation
            self.log(
                f"val/{k}",
                v,
                on_step=False,
                on_epoch=True,
                batch_size=batch["views"][0].shape[0],
            )
        return loss["objective"]

    def test_step(self, batch: Dict[str, Any], batch_idx: int) -> torch.Tensor:
        """Performs one step of testing on a batch of views."""
        loss = self.loss(batch)
        for k, v in loss.items():
            # Use f-string instead of concatenation
            self.log(
                f"test/{k}",
                v,
                on_step=False,
                on_epoch=True,
                batch_size=batch["views"][0].shape[0],
            )
        return loss["objective"]

    @torch.no_grad()
    def get_representations(
        self,
        loader: torch.utils.data.DataLoader,
    ):
        self.eval()  # Ensure the model is in evaluation mode
        all_z = []

        for batch in loader:
            views_device = [view.to(self.device) for view in batch["views"]]
            z = self(views_device)
            all_z.append([z_.cpu() for z_ in z])

        # Stack all latent vectors along dimension 0 (batches)
        stacked_z = [torch.vstack(z_) for z_ in zip(*all_z)]

        return stacked_z

    @torch.no_grad()
    def transform(
        self,
        loader: torch.utils.data.DataLoader,
    ) -> List[np.ndarray]:
        """Returns the latent representations for each view in the loader."""
        return [z.numpy() for z in self.get_representations(loader)]

    def configure_optimizers(
        self,
    ) -> Union[
        torch.optim.Optimizer,
        Tuple[List[torch.optim.Optimizer], List[torch.optim.lr_scheduler._LRScheduler]],
    ]:
        """Configures the optimizer and the learning rate scheduler."""
        if self.optimizer == "sgd":
            optimizer = torch.optim.SGD
        elif self.optimizer == "adam":
            optimizer = torch.optim.Adam
        elif self.optimizer == "adamw":
            optimizer = torch.optim.AdamW
        elif self.optimizer == "lbfgs":
            optimizer = torch.optim.LBFGS
        else:
            raise ValueError(f"{self.optimizer} not in (sgd, adam, adamw)")

        # create optimizer
        optimizer = optimizer(
            self.parameters(),
            lr=self.lr,
            **self.extra_optimizer_kwargs,
        )
        if self.scheduler is None:
            return optimizer
        elif self.scheduler == "cosine":
            scheduler = CosineAnnealingLR(
                optimizer, self.max_epochs, eta_min=self.min_lr
            )
        elif self.scheduler == "step":
            scheduler = MultiStepLR(optimizer, self.lr_decay_steps)
        else:
            raise ValueError(f"{self.scheduler} not in (warmup_cosine, cosine, step)")
        return [optimizer], [scheduler]

    def configure_callbacks(self) -> None:
        """Configures the callbacks for the model."""
        pass

    @staticmethod
    def detach_all(z: List[torch.Tensor]) -> List[torch.Tensor]:
        """Detaches all tensors in a list from the computation graph."""
        # Use list comprehension instead of for loop
        return [z_.detach() for z_ in z]



import seaborn as sns

from cca_zoo.utils.check_values import check_tsne_support, check_seaborn_support
from cca_zoo.visualisation import ScoreDisplay


class TSNEScoreDisplay(ScoreDisplay):
    def _validate_plot_params(self):
        check_tsne_support("TSNEScoreDisplay")
        check_seaborn_support("TSNEScoreDisplay")

    def plot(self):
        self._validate_plot_params()
        import openTSNE
        import matplotlib.pyplot as plt

        reducer = openTSNE.TSNE()
        embedding = reducer.fit(self.train_scores[0])
        fig, ax = plt.subplots()
        sns.scatterplot(
            x=embedding[:, 0],
            y=embedding[:, 1],
            hue=self.train_labels,
            ax=ax,
            alpha=0.1 if self.test_scores is not None else 1.0,
            label="Train" if self.test_scores is not None else None,
            **self.kwargs,
        )
        if self.test_scores is not None:
            embedding = reducer.fit(self.test_scores[0])
            sns.scatterplot(
                x=embedding[:, 0],
                y=embedding[:, 1],
                hue=self.test_labels,
                ax=ax,
                label="Test",
                **self.kwargs,
            )
        plt.tight_layout()
        self.figure_ = fig
        return self



import torch

from ._dcca_ey import DCCA_EY


class DCCA_GHA(DCCA_EY):
    def get_AB(self, z):
        A = torch.zeros(
            self.latent_dimensions, self.latent_dimensions, device=z[0].device
        )  # initialize the cross-covariance matrix
        B = torch.zeros(
            self.latent_dimensions, self.latent_dimensions, device=z[0].device
        )  # initialize the auto-covariance matrix
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    B += torch.cov(zi.T)  # add the auto-covariance of each view to B
                A += torch.cov(torch.hstack((zi, zj)).T)[
                    self.latent_dimensions :, : self.latent_dimensions
                ]  # add the cross-covariance of each pair of views to A
        return A / len(z), B / len(
            z
        )  # return the normalized matrices (divided by the number of views)

    def loss(self, batch, **kwargs):
        z = self(batch["views"])
        independent_views = batch.get("independent_views", None)
        A, B = self.get_AB(z)
        rewards = torch.trace(2 * A)
        if independent_views is None:
            # Hebbian
            penalties = torch.trace(A.detach() @ B)
            # penalties = torch.trace(A @ B)
        else:
            independent_z = self(independent_views)
            independent_A, independent_B = self.get_AB(independent_z)
            # Hebbian
            penalties = torch.trace(independent_A.detach() @ B)
            # penalties = torch.trace(A @ independent_B)
        return {
            "objective": -rewards.sum() + penalties,
            "rewards": rewards.sum(),
            "penalties": penalties,
        }



import torch

from cca_zoo.linear._gradient._ey import CCA_EY


class CCA_GHA(CCA_EY):
    def _more_tags(self):
        return {"multiview": True, "stochastic": True}

    def get_AB(self, z):
        latent_dims = z[0].shape[1]
        A = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the cross-covariance matrix
        B = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the auto-covariance matrix
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    B += self._cross_covariance(zi, zj, latent_dims)
                A += self._cross_covariance(zi, zj, latent_dims)
        return A / len(z), B / len(z)

    def loss(self, views, independent_views=None, **kwargs):
        # Encoding the views with the forward method
        z = self(views)
        # Getting A and B matrices from z
        A, B = self.get_AB(z)
        rewards = torch.trace(2 * A)
        if independent_views is None:
            # Hebbian
            penalties = torch.trace(A.detach() @ B)
            # penalties = torch.trace(A @ B)
        else:
            # Encoding another set of views with the forward method
            independent_z = self(independent_views)
            # Getting A' and B' matrices from independent_z
            independent_A, independent_B = self.get_AB(independent_z)
            # Hebbian
            penalties = torch.trace(independent_A.detach() @ B)
            # penalties = torch.trace(A @ independent_B)
        return {
            "loss": -rewards + penalties,
            "rewards": rewards,
            "penalties": penalties,
        }



from typing import Iterable, Union

import numpy as np
from scipy.linalg import block_diag, eigh
from sklearn.decomposition import PCA

from cca_zoo._base import BaseModel
from cca_zoo.utils import _process_parameter


class MCCA(BaseModel):
    r"""
    A class used to fit a Regularised CCA (canonical ridge) model. This model adds a regularization term to the CCA objective function to avoid overfitting and improve stability. It uses PCA to perform the optimization efficiently for high dimensional data.

    The objective function of regularised CCA is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        (1-c_1)w_1^TX_1^TX_1w_1+c_1w_1^Tw_1=n

        (1-c_2)w_2^TX_2^TX_2w_2+c_2w_2^Tw_2=n

    where :math:`c_i` are the regularization parameters for each view.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state, by default None
    c : Union[Iterable[float], float], optional
        Regularisation parameter, by default None
    accept_sparse : Union[bool, str], optional
        Whether to accept sparse data, by default None

    Examples
    --------
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = MCCA()
    >>> model.fit((X1,X2)).score((X1,X2))

    References
    --------
    Vinod, Hrishikesh _D. "Canonical ridge and econometrics of joint production." Journal of econometrics 4.2 (1976): 147-166.
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        c: Union[Iterable[float], float] = None,
        accept_sparse=None,
        eps: float = 1e-6,
        pca: bool = True,
    ):
        # Set the default value for accept_sparse
        if accept_sparse is None:
            accept_sparse = ["csc", "csr"]
        # Call the parent class constructor
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
        )
        # Store the c parameter
        self.c = c
        self.eps = eps
        self.pca = pca

    def _check_params(self):
        # Process the c parameter for each view
        self.c = _process_parameter("c", self.c, 0, self.n_views_)

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        # Validate the input data
        views = self._validate_data(views)
        # Check the parameters
        self._check_params()
        views = self._process_data(views, **kwargs)
        eigvals, eigvecs = self._solve_gevp(views, y=y, **kwargs)
        # Compute the weights for each view
        self._weights(eigvals, eigvecs, views, **kwargs)
        # delete pca to save memory
        if self.pca:
            del self.pca_models
        return self

    def _process_data(self, views, **kwargs):
        if self.pca:
            views = self._apply_pca(views)
        return views

    def _solve_gevp(self, views: Iterable[np.ndarray], y=None, **kwargs):
        # Setup the eigenvalue problem
        C = self._C(views, **kwargs)
        D = self._D(views, **kwargs)
        self.splits = np.cumsum([view.shape[1] for view in views])
        # Solve the eigenvalue problem
        # Get the dimension of _C
        p = C.shape[0]
        # Solve the generalized eigenvalue problem Cx=lambda Dx using a subset of eigenvalues and eigenvectors
        [eigvals, eigvecs] = eigh(
            C,
            D,
            subset_by_index=[p - self.latent_dimensions, p - 1],
        )
        # Sort the eigenvalues and eigenvectors in descending order
        idx = np.argsort(eigvals, axis=0)[::-1]
        eigvecs = eigvecs[:, idx].real
        return np.flip(eigvals), eigvecs

    def _weights(self, eigvals, eigvecs, views, **kwargs):
        # split eigvecs into weights for each view
        self.weights = np.split(eigvecs, self.splits[:-1], axis=0)
        if self.pca:
            # go from weights in PCA space to weights in original space
            self.weights = [
                pca.components_.T @ self.weights[i]
                for i, pca in enumerate(self.pca_models)
            ]

    def _apply_pca(self, views):
        """
        Do data driven PCA on each view
        """
        self.pca_models = [PCA() for _ in views]
        # Fit PCA on each view
        return [self.pca_models[i].fit_transform(view) for i, view in enumerate(views)]

    def _C(self, views, **kwargs):
        all_views = np.hstack(views)
        C = np.cov(all_views, rowvar=False)
        C -= block_diag(*[np.cov(view, rowvar=False) for view in views])
        return C / len(views)

    def _D(self, views, **kwargs):
        if self.pca:
            # Can regularise by adding to diagonal
            D = block_diag(
                *[
                    np.diag((1 - self.c[i]) * pc.explained_variance_ + self.c[i])
                    for i, pc in enumerate(self.pca_models)
                ]
            )
        else:
            D = block_diag(
                *[
                    (1 - self.c[i]) * np.cov(view, rowvar=False)
                    + self.c[i] * np.eye(view.shape[1])
                    for i, view in enumerate(views)
                ]
            )
        D_smallest_eig = min(0, np.linalg.eigvalsh(D).min()) - self.eps
        D = D - D_smallest_eig * np.eye(D.shape[0])
        return D / len(views)

    def _more_tags(self):
        # Indicate that this class is for multiview data
        return {"multiview": True}


class rCCA(MCCA):
    r"""
    A class used to fit Regularised CCA (canonical ridge) model. This model adds a regularization term to the CCA objective function to avoid overfitting and improve stability. It uses PCA to perform the optimization efficiently for high dimensional data.

    The objective function of regularised CCA is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        (1-c_1)w_1^TX_1^TX_1w_1+c_1w_1^Tw_1=n

        (1-c_2)w_2^TX_2^TX_2w_2+c_2w_2^Tw_2=n

    where :math:`c_i` are the regularization parameters for each view.

    Examples
    --------
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = rCCA(c=0.1)
    >>> model.fit((X1,X2)).score((X1,X2))


    References
    --------
    Vinod, Hrishikesh _D. "Canonical ridge and econometrics of joint production." Journal of econometrics 4.2 (1976): 147-166.
    """

    def _C(self, views, **kwargs):
        if len(views) != 2:
            raise ValueError(
                f"Model can only be used with two views, but {len(views)} were given. Use MCCA or GCCA instead for CCA or MPLS for PLS."
            )
        # Compute the B matrices for each view
        B = [
            (1 - self.c[i]) * pc.explained_variance_ + self.c[i]
            for i, pc in enumerate(self.pca_models)
        ]
        C = np.cov(views[0] / np.sqrt(B[0]), views[1] / np.sqrt(B[1]), rowvar=False)[
            0 : views[0].shape[1], views[0].shape[1] :
        ]
        # if views[0].shape[1] <= views[1].shape[1] then return R@R^T else return R^T@R
        if views[0].shape[1] <= views[1].shape[1]:
            self.primary_view = 0
            return C @ C.T
        else:
            self.primary_view = 1
            return C.T @ C

    def _D(self, views, **kwargs):
        return None

    def _weights(self, eigvals, eigvecs, views):
        B = [
            (1 - self.c[i]) * pc.singular_values_**2 / self.n_samples_ + self.c[i]
            for i, pc in enumerate(self.pca_models)
        ]
        C = np.cov(
            views[self.primary_view], views[1 - self.primary_view], rowvar=False
        )[0 : views[self.primary_view].shape[1], views[self.primary_view].shape[1] :]
        self.weights = [None] * 2
        # Compute the weight matrix for primary view
        self.weights[1 - self.primary_view] = (
            # Project view 1 onto its principal components
            self.pca_models[1 - self.primary_view].components_.T
            # Scale by the inverse of B[0]
            @ np.diag(1 / B[1 - self.primary_view])
            # Multiply by the cross-covariance matrix
            @ C.T
            # Scale by the inverse of the square root of B[1]
            @ np.diag(1 / np.sqrt(B[self.primary_view]))
            # Multiply by the eigenvectors
            @ eigvecs
            # Scale by the inverse of the square root of eigenvalues
            / np.sqrt(eigvals)
        )

        # Compute the weight matrix for view 2
        self.weights[self.primary_view] = (
            # Project view 2 onto its principal components
            self.pca_models[self.primary_view].components_.T
            # Scale by the inverse of the square root of B[1]
            @ np.diag(1 / np.sqrt(B[self.primary_view]))
            # Multiply by the eigenvectors
            @ eigvecs
        )


class CCA(rCCA):
    r"""
    A class used to fit a simple CCA model. This model finds the linear projections of two views that maximize their correlation.

    The objective function of CCA is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        w_1^TX_1^TX_1w_1=n

        w_2^TX_2^TX_2w_2=n

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random seed for reproducibility, by default None

    References
    --------

    Hotelling, Harold. "Relations between two sets of variates." Breakthroughs in statistics. Springer, New York, NY, 1992. 162-190.

    Example
    -------
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = CCA()
    >>> model.fit((X1,X2)).score((X1,X2))
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
    ):
        # Call the parent class constructor with c=0.0 to disable regularization
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            c=0.0,
            random_state=random_state,
        )



from ._search import GridSearchCV, RandomizedSearchCV
from ._validation import cross_validate, learning_curve, permutation_test_score

__all__ = [
    "GridSearchCV",
    "RandomizedSearchCV",
    "cross_validate",
    "learning_curve",
    "permutation_test_score",
]



from ._gcca import GCCA
from ._gradient import CCA_EY, CCA_GHA, CCA_SVD, PLS_EY, PLSStochasticPower
from ._grcca import GRCCA
from ._iterative import (
    PLS_ALS,
    SCCA_IPLS,
    SCCA_PMD,
    # AltMaxVar,
    ElasticCCA,
    SCCA_Parkhomenko,
    SCCA_Span,
)
from ._mcca import CCA, MCCA, rCCA
from ._partialcca import PartialCCA
from ._pcacca import PCACCA
from ._pls import MPLS, PLS
from ._prcca import PRCCA
from ._tcca import TCCA

__all__ = [
    "MCCA",
    "CCA",
    "rCCA",
    "PLS",
    "MPLS",
    "GCCA",
    "GRCCA",
    "PartialCCA",
    "PRCCA",
    "TCCA",
    "PCACCA",
    # "AltMaxVar",
    "SCCA_IPLS",
    "ElasticCCA",
    "PLS_ALS",
    "SCCA_PMD",
    "SCCA_Parkhomenko",
    "SCCA_Span",
    "CCA_EY",
    "PLS_EY",
    "CCA_GHA",
    "CCA_SVD",
    "PLSStochasticPower",
]



"""
Class which allows for the different (or the same) processing of multiple views of data.
"""
from mvlearn.utils import check_Xs
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_is_fitted


class MultiViewPreprocessing(TransformerMixin):
    def __init__(self, preprocessing_list):
        self.preprocessing_list = preprocessing_list

    def fit(self, views, y=None):
        """
        Fits the associated preprocessing steps to each view.
        Parameters
        ----------
        views
        y

        Returns
        -------

        """
        if len(self.preprocessing_list) == 1:
            self.preprocessing_list = self.preprocessing_list * len(views)
        elif len(self.preprocessing_list) != len(views):
            raise ValueError(
                "Length of preprocessing_list must be 1 (apply the same preprocessing to each view) or equal to the number of views"
            )
        check_Xs(views, enforce_views=range(len(self.preprocessing_list)))
        for view, preprocessing in zip(views, self.preprocessing_list):
            # Skip if preprocessing is None
            if preprocessing is not None:
                preprocessing.fit(view, y)
        return self

    def transform(self, X, y=None):
        """
        Transforms each view using the associated preprocessing steps.
        Parameters
        ----------
        X
        y

        Returns
        -------

        """
        [
            check_is_fitted(preprocessing)
            for preprocessing in self.preprocessing_list
            if preprocessing is not None
        ]
        check_Xs(X, enforce_views=range(len(self.preprocessing_list)))
        return [
            # Skip if preprocessing is None
            view if preprocessing is None else preprocessing.transform(view)
            for view, preprocessing in zip(X, self.preprocessing_list)
        ]



import matplotlib.pyplot as plt
import seaborn as sns


class WeightHeatmapDisplay:
    """Heatmap of the weights of a model.

    Parameters
    ----------
    model : CCA model
        A fitted CCA model.
    """

    def __init__(self, weights, view_labels=None, **kwargs):
        self.weights = weights
        self.view_labels = view_labels
        self.kwargs = kwargs

    @classmethod
    def from_estimator(cls, model, view_labels=None, **kwargs):
        weights = model.weights
        return cls.from_weights(weights, view_labels=view_labels, **kwargs)

    @classmethod
    def from_weights(cls, weights, view_labels=None, **kwargs):
        return cls(weights, view_labels=view_labels, **kwargs)

    def plot(self, **kwargs):
        """Plot the heatmap.

        Parameters
        ----------
        ax : matplotlib axes, optional
            Axes to plot on, by default None.
        kwargs : dict
            Keyword arguments to pass to seaborn.heatmap

        Returns
        -------
        ax : matplotlib axes
            Axes with the heatmap.
        """
        fig, axs = plt.subplots(1, len(self.weights), figsize=(10, 5))
        if self.view_labels is None:
            self.view_labels = [f"View {i}" for i in range(len(self.weights))]
        self.weights_cov = [w.T @ w for w in self.weights]
        # loop through each view and have a heatmap of the covariance of the weights
        for i, view_weights_cov in enumerate(self.weights_cov):
            sns.heatmap(view_weights_cov, ax=axs[i], **self.kwargs)
            axs[i].set_title(self.view_labels[i])
        plt.tight_layout()
        self.figure_ = fig
        return self



# from typing import Iterable, Union
#
# import numpy as np
# import torch
#
# from cca_zoo.linear._iterative._base import BaseIterative, BaseLoop
# from cca_zoo.utils import _process_parameter
#
# #
# class AltMaxVar(BaseIterative):
#     def __init__(
#         self,
#         latent_dimensions=1,
#         copy_data=True,
#         random_state=None,
#         epochs=100,
#         tol=1e-3,
#         proximal="L1",
#         positive=False,
#         initialization="uniform",
#         tau: Union[Iterable[float], float] = None,
#         proximal_params: Iterable[dict] = None,
#         gamma=0.1,
#         learning_rate=1e-2,
#         T=100,
#         trainer_kwargs=None,
#         convergence_checking=None,
#         track=None,
#         verbose=False,
#     ):
#         super().__init__(
#             latent_dimensions=latent_dimensions,
#             copy_data=copy_data,
#             random_state=random_state,
#             tol=tol,
#             epochs=epochs,
#             convergence_checking=convergence_checking,
#             track=track,
#             verbose=verbose,
#             trainer_kwargs=trainer_kwargs,
#             initialization=initialization,
#         )
#         self.tau = tau
#         self.proximal = proximal
#         self.proximal_params = proximal_params
#         self.gamma = gamma
#         self.learning_rate = learning_rate
#         self.T = T
#         self.positive = positive
#         # set trainer kwargs accelerator to 'cpu'
#         self.trainer_kwargs["accelerator"] = "cpu"
#
#     def _get_pl_module(self, weights=None, k=None):
#         return AltMaxVarLoop(
#             weights=weights,
#             k=k,
#             gamma=self.gamma,
#             T=self.T,
#             proximal_operators=self.proximal_operators,
#             learning_rate=self.learning_rate,
#             convergence_checking=self.convergence_checking,
#             track=self.track,
#         )
#
#     def _check_params(self):
#         self.proximal = _process_parameter(
#             "proximal", self.proximal, "L1", self.n_views_
#         )
#         self.positive = _process_parameter(
#             "positive", self.positive, False, self.n_views_
#         )
#         self.tau = _process_parameter("tau", self.tau, 0, self.n_views_)
#         self.sigma = self.tau
#         self.proximal_operators = [
#             self._get_proximal(view) for view in range(self.n_views_)
#         ]
#
#     def _get_proximal(self, view):
#         if callable(self.proximal[view]):
#             params = self.proximal_params[view] or {}
#         else:
#             params = {
#                 "sigma": self.sigma[view],
#                 "positive": self.positive[view],
#             }
#         return _proximal_operators(self.proximal[view], **params)
#
#     def _more_tags(self):
#         return {"multiview": True}
#
#
# class AltMaxVarLoop(BaseLoop):
#     def __init__(
#         self,
#         weights,
#         k=None,
#         gamma=0.1,
#         T=100,
#         proximal_operators=None,
#         learning_rate=1e-3,
#         convergence_checking=None,
#         track=None,
#     ):
#         super().__init__(
#             weights=weights,
#             k=k,
#             convergence_checking=convergence_checking,
#             tracking=track,
#         )
#         self.gamma = gamma
#         self.proximal_operators = proximal_operators
#         self.T = T
#         self.learning_rate = learning_rate
#
#     def forward(self, views: list) -> list:
#         # views detach and numpy
#         views = [view.detach().numpy() for view in views]
#         return [view @ weight for view, weight in zip(views, self.weights)]
#
#     def _get_target(self, scores):
#         if hasattr(self, "G"):
#             R = self.gamma * scores.mean(axis=0) + (1 - self.gamma) * self.G
#         else:
#             R = scores.mean(axis=0)
#         U, S, Vt = np.linalg.svd(R, full_matrices=False)
#         G = U @ Vt
#         return G / np.sqrt(np.diag(np.atleast_1d(np.cov(G, rowvar=False))))
#
#     def objective(self, views, scores, weights) -> int:
#         least_squares = (np.linalg.norm(scores - self.G, axis=(1, 2)) ** 2).sum()
#         regularization = np.array(
#             [self.proximal_operators[view](weights[view]) for view in range(len(views))]
#         ).sum()
#         return least_squares + regularization
#
#     def training_step(self, batch, batch_idx):
#         scores = np.stack(self(batch["views"]))
#         self.G = self._get_target(scores)
#         old_weights = self.weights.copy()
#         for i, view in enumerate(batch["views"]):
#             view = view.detach().numpy()
#             t = 0
#             prev_weights = None
#             converged = False
#             while t < self.T and not converged:
#                 grad = view.T @ (view @ self.weights[i] - self.G) / view.shape[0]
#                 # update the weights using the gradient descent and proximal operator
#                 self.weights[i] -= self.learning_rate * grad
#                 self.weights[i] = self.proximal_operators[i].prox(
#                     self.weights[i], self.learning_rate
#                 )
#                 # check if the weights have changed significantly from the previous iteration
#                 if prev_weights is not None and np.allclose(
#                     self.weights[i], prev_weights
#                 ):
#                     # if yes, set converged to True and break the loop
#                     converged = True
#                 # update the previous weights for the next iteration
#                 prev_weights = self.weights[i].copy()
#                 t += 1
#
#         # if track or convergence_checking is enabled, compute the objective function
#         if self.tracking or self.convergence_checking:
#             objective = self.objective(batch["views"], scores, self.weights)
#             # check that the maximum change in weights is smaller than the tolerance times the maximum absolute value of the weights
#             weights_change = torch.tensor(
#                 np.max(
#                     [
#                         np.max(np.abs(old_weights[i] - self.weights[i]))
#                         / np.max(np.abs(self.weights[i]))
#                         for i in range(len(self.weights))
#                     ]
#                 )
#             )
#             return {"loss": torch.tensor(objective), "weights_change": weights_change}
#
#
# from pyproximal import (
#     L0,
#     L1,
#     L2,
#     L21,
#     Euclidean,
#     EuclideanBall,
#     L0Ball,
#     L1Ball,
#     L21_plus_L1,
#     Log,
#     Log1,
#     Nuclear,
#     NuclearBall,
# )
#
# PROXIMAL_OPERATORS = {
#     "L0": L0,
#     "L0Ball": L0Ball,
#     "L1": L1,
#     "L1Ball": L1Ball,
#     "L2": L2,
#     "L21": L21,
#     "L21_plus_L1": L21_plus_L1,
#     "Nuclear": Nuclear,
#     "NuclearBall": NuclearBall,
#     "Log": Log,
#     "Log1": Log1,
#     "Euclidean": Euclidean,
#     "EuclideanBall": EuclideanBall,
# }
#
# PROXIMAL_PARAMS = {
#     "Dummy": (),
#     "L0": frozenset(["sigma"]),
#     "L0Ball": frozenset(["radius"]),
#     "L1": frozenset(["sigma"]),
#     "L1Ball": frozenset(["n", "radius"]),
#     "L2": frozenset(["sigma"]),
#     "L21": frozenset(["ndim", "sigma"]),
#     "L21_plus_L1": frozenset(["sigma", "rho"]),
#     "TV": frozenset(["sigma", "isotropic", "dims"]),
#     "Nuclear": frozenset(["dim", "sigma"]),
#     "NuclearBall": frozenset(["dims", "radius"]),
#     "Log": frozenset(["sigma", "gamma"]),
#     "Log1": frozenset(["sigma", "delta"]),
#     "Euclidean": frozenset(["sigma"]),
#     "TVL1": frozenset(["sigma", "shape", "l1_ratio"]),
# }
#
#
# def _proximal_operators(proximal, filter_params=True, **params):
#     if proximal in PROXIMAL_OPERATORS:
#         if filter_params:
#             params = {k: params[k] for k in params if k in PROXIMAL_PARAMS[proximal]}
#         return PROXIMAL_OPERATORS[proximal](**params)
#     elif callable(proximal):
#         return proximal(**params)



"""
Module for finding CCA effects sequentially by deflation.

Check if each effect is significant, and if so, remove it from the data and repeat.
"""
from abc import ABCMeta
from typing import Iterable

import numpy as np
from sklearn.base import MetaEstimatorMixin

from cca_zoo._base import BaseModel
from cca_zoo.linear._iterative._deflation import deflate_views
from cca_zoo.model_selection._validation import permutation_test_score


class SequentialModel(MetaEstimatorMixin, BaseModel, metaclass=ABCMeta):
    def __init__(
        self,
        estimator,
        estimator_hyperparams=None,
        permutation_test_params=None,
        latent_dimensions=None,  # Maximum number of latent dimensions to fit
        copy_data=True,
        accept_sparse=False,
        random_state=None,
        permutation_test=False,  # Whether to use permutation test to determine significance
        p_threshold=1e-3,  # Threshold for permutation test if used
        corr_threshold=0.0,  # Threshold for effect size if permutation test not used
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
        )
        # Check the estimator has 1 latent dimension or if it is GridSearchCV or RandomizedSearchCV that the base estimator has 1 latent dimension
        if hasattr(estimator, "estimator"):
            if estimator.estimator.latent_dimensions != 1:
                raise ValueError(
                    "The estimator must have 1 latent dimension, but has {}".format(
                        estimator.estimator.latent_dimensions
                    )
                )
        elif estimator.latent_dimensions != 1:
            raise ValueError(
                "The estimator must have 1 latent dimension, but has {}".format(
                    estimator.latent_dimensions
                )
            )
        self.estimator = estimator
        if estimator_hyperparams is None:
            estimator_hyperparams = {}
        self.estimator_hyperparams = estimator_hyperparams
        self.permutation_test = permutation_test
        if permutation_test_params is None:
            permutation_test_params = {}
        self.permutation_test_params = permutation_test_params
        self.p_threshold = p_threshold
        self.corr_threshold = corr_threshold

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        # Validate the input data and parameters
        self._validate_data(views)
        self._check_params()
        # Set the default latent dimensions to the minimum number of features
        if self.latent_dimensions is None:
            self.latent_dimensions = min([view.shape[1] for view in views])
        # Initialize the weights and p-values lists
        self.weights = [[] for view in views]
        self.p_values = []
        # Loop over the latent dimensions
        for k in range(self.latent_dimensions):
            # Fit the estimator with the current views
            self.estimator.set_params(**self.estimator_hyperparams)
            self.estimator.fit(views)
            # Perform permutation test if required
            if self.permutation_test:
                # Get the best estimator if it exists, otherwise use the original estimator
                best_estimator = getattr(
                    self.estimator, "best_estimator_", self.estimator
                )
                # Get the p-value from the permutation test score
                p_value = permutation_test_score(
                    best_estimator,
                    views,
                    y=None,
                    **self.permutation_test_params,
                )[2]
                # Append the p-value to the list
                self.p_values.append(p_value)
            # Check if the stopping criterion is met based on p-value or correlation score
            if (
                p_value >= self.p_threshold
                or best_estimator.score(views) < self.corr_threshold
            ):
                self.p_values.pop()
                break
            else:
                # Deflate the views and store the weights
                views = deflate_views(views, best_estimator.weights)
                for i, weight in enumerate(best_estimator.weights):
                    self.weights[i].append(weight)
        # Set the final latent dimensions to k
        self.latent_dimensions = k
        # Concatenate the weights from each effect
        self.weights = [np.concatenate(weights, axis=1) for weights in self.weights]
        return self



import tensorly as tl
import torch
from tensorly.cp_tensor import cp_to_tensor
from tensorly.decomposition import parafac


def inv_sqrtm(A, eps=1e-9):
    """Compute the inverse square-root of a positive definite matrix."""
    # Perform eigendecomposition of covariance matrix
    U, S, V = torch.svd(A)
    # Enforce positive definite by taking a torch max() with eps
    S = torch.max(S, torch.tensor(eps, device=S.device))
    # Calculate inverse square-root
    inv_sqrt_S = torch.diag_embed(torch.pow(S, -0.5))
    # Calculate inverse square-root matrix
    B = torch.matmul(torch.matmul(U, inv_sqrt_S), V.transpose(-1, -2))
    return B


def _demean(views):
    return tuple([view - view.mean(dim=0) for view in views])


class MCCA:
    """Differentiable MCCA Loss. Solves the multiset eigenvalue problem.

    References
    ----------
    https://arxiv.org/pdf/2005.11914.pdf

    """

    def __init__(self, r: float = 0, eps: float = 1e-3):
        self.r = r
        self.eps = eps

    def C(self, views):
        """Calculate cross-covariance matrix."""
        all_views = torch.cat(views, dim=1)
        C = torch.cov(all_views.T)
        C = C - torch.block_diag(*[torch.cov(view.T) for view in views])
        return C / len(views)

    def D(self, views):
        """Calculate block covariance matrix."""
        D = torch.block_diag(
            *[
                (1 - self.r) * torch.cov(view.T)
                + self.r * torch.eye(view.shape[1], device=view.device)
                for view in views
            ]
        )
        return D / len(views)

    def correlation(self, views):
        """Calculate correlation."""
        latent_dims = views[0].shape[1]
        views = _demean(views)
        C = self.C(views)
        D = self.D(views)
        C += D
        R = inv_sqrtm(D, self.eps)
        C_whitened = R @ C @ R.T
        eigvals = torch.linalg.eigvalsh(C_whitened)
        idx = torch.argsort(eigvals, descending=True)
        eigvals = eigvals[idx[:latent_dims]]
        return eigvals

    def loss(self, views):
        """Calculate loss."""
        eigvals = self.correlation(views)
        eigvals = torch.nn.LeakyReLU()(eigvals[torch.gt(eigvals, 0)])
        corr = eigvals.sum()
        return -corr


class GCCA:
    """Differentiable GCCA Loss. Solves the generalized CCA eigenproblem.

    References
    ----------
    https://arxiv.org/pdf/2005.11914.pdf
    """

    def __init__(self, r: float = 0, eps: float = 1e-3):
        self.r = r
        self.eps = eps

    def Q(self, views):
        """Calculate Q matrix."""
        eigen_views = [
            view @ torch.linalg.inv(torch.cov(view.T)) @ view.T for view in views
        ]
        Q = torch.stack(eigen_views, dim=0).sum(dim=0)
        return Q

    def correlation(self, views):
        """Calculate correlation."""
        latent_dims = views[0].shape[1]
        views = _demean(views)
        Q = self.Q(views)
        eigvals = torch.linalg.eigvalsh(Q)
        idx = torch.argsort(eigvals, descending=True)
        eigvals = eigvals[idx[:latent_dims]]
        return torch.nn.LeakyReLU()(eigvals)

    def loss(self, views):
        """Calculate loss."""
        eigvals = self.correlation(views)
        corr = eigvals.sum()
        return -corr


# Original work Copyright (c) 2016 Vahid Noroozi
# Modified work Copyright 2019 Zhanghao Wu

# Permission is hereby granted, free of chviewe, to any person obtaining a copy
# of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:


# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
class CCA:
    """Differentiable CCA Loss. Solves the CCA problem."""

    def __init__(self, r: float = 0, eps: float = 1e-3):
        self.r = r
        self.eps = eps

    def correlation(self, views):
        """Calculate correlation."""
        latent_dims = views[0].shape[1]
        o1 = views[0].shape[1]
        o2 = views[1].shape[1]

        views = _demean(views)

        SigmaHat12 = torch.cov(torch.hstack((views[0], views[1])).T)[
            :latent_dims, latent_dims:
        ]
        SigmaHat11 = torch.cov(views[0].T) + self.r * torch.eye(
            o1, device=views[0].device
        )
        SigmaHat22 = torch.cov(views[1].T) + self.r * torch.eye(
            o2, device=views[1].device
        )

        SigmaHat11RootInv = inv_sqrtm(SigmaHat11, self.eps)
        SigmaHat22RootInv = inv_sqrtm(SigmaHat22, self.eps)

        Tval = SigmaHat11RootInv @ SigmaHat12 @ SigmaHat22RootInv
        trace_TT = Tval.T @ Tval
        eigvals = torch.linalg.eigvalsh(trace_TT)

        return eigvals

    def loss(self, views):
        """Calculate loss."""
        eigvals = self.correlation(views)
        eigvals = torch.nn.LeakyReLU()(eigvals[torch.gt(eigvals, 0)])
        return -eigvals.sum()


class TCCA:
    """Differentiable TCCA Loss."""

    def __init__(self, r: float = 0, eps: float = 1e-4):
        self.r = r
        self.eps = eps

    def loss(self, views):
        latent_dims = views[0].shape[1]
        views = _demean(views)
        covs = [
            (1 - self.r) * torch.cov(view.T)
            + self.r * torch.eye(view.size(1), device=view.device)
            for view in views
        ]
        whitened_z = [view @ inv_sqrtm(cov, self.eps) for view, cov in zip(views, covs)]
        # The idea here is to form a matrix with M dimensions one for each view where at index
        # M[p_i,p_j,p_k...] we have the sum over n samples of the product of the pth feature of the
        # ith, jth, kth view etc.
        for i, el in enumerate(whitened_z):
            # To achieve this we start with the first view so M is nxp.
            if i == 0:
                M = el
            # For the remaining views we expand their dimensions to match M i.e. nx1x...x1xp
            else:
                for _ in range(len(M.size()) - 1):
                    el = torch.unsqueeze(el, 1)
                # Then we perform an outer product by expanding the dimensionality of M and
                # outer product with the expanded el
                M = torch.unsqueeze(M, -1) @ el
        M = torch.mean(M, 0)
        tl.set_backend("pytorch")
        M_parafac = parafac(
            M.detach(), latent_dims, verbose=False, normalize_factors=True
        )
        M_parafac.weights = 1
        M_hat = cp_to_tensor(M_parafac)
        return torch.linalg.norm(M - M_hat)



from . import (
    data,
    deep,
    linear,
    model_selection,
    visualisation,
    preprocessing,
    sequential,
)

__all__ = [
    "data",
    "model_selection",
    "linear",
    "visualisation",
    "deep",
]
try:
    from . import probabilistic

    __all__.append("probabilistic")
except ModuleNotFoundError:
    pass



from typing import Union, Iterable

import numpy as np
from sklearn.linear_model import ElasticNet, Lasso, Ridge, SGDRegressor

from cca_zoo.linear._iterative._base import BaseIterative
from cca_zoo.linear._iterative._deflation import DeflationMixin
from cca_zoo.utils import _process_parameter


# Define SCCA_Elastic class
class ElasticCCA(DeflationMixin, BaseIterative):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        accept_sparse=None,
        epochs=100,
        initialization: Union[str, callable] = "uniform",
        early_stopping=False,
        verbose=None,
        alpha=None,  # regularization parameter for Elastic
        l1_ratio=None,  # ratio of L1 to L2 penalty for Elastic
        positive=None,  # whether to enforce positive coefficients for Elastic
        stochastic=False,  # whether to use stochastic gradient descent for Elastic
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            tol=tol,
            accept_sparse=accept_sparse,
            epochs=epochs,
            initialization=initialization,
            early_stopping=early_stopping,
            verbose=verbose,
        )
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.positive = positive
        self.stochastic = stochastic

    def _check_params(self):
        self.alpha = _process_parameter("alpha", self.alpha, 0, self.n_views_)
        self.l1_ratio = _process_parameter("l1_ratio", self.l1_ratio, 0, self.n_views_)
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views_
        )
        self.regressors = initialize_regressors(
            self.alpha,
            self.l1_ratio,
            self.positive,
            self.stochastic,
            self.tol,
            self.random_state,
        )

    def _update_weights(self, views: Iterable[np.ndarray], i: int):
        # Update the weights for the current view using Elastic
        # Get the scores of all views
        scores = np.stack(self.transform(views))
        # Compute the target by summing the scores along dim 0 and dividing by the square root of the covariance of the target
        target = np.sum(scores, axis=0)
        target = target / np.linalg.norm(target)

        # Loop over the views and fit each regressor to the view and the target
        self.regressors[i] = self.regressors[i].fit(views[i], target)
        # Update the weights with the coefficients of each regressor
        new_weights = np.squeeze(self.regressors[i].coef_)
        # Return the updated weights
        return new_weights[:, None]

    def _objective(self, views: Iterable[np.ndarray]):
        scores = np.stack(self.transform(views))
        objective = 0
        for view_index, view in enumerate(views):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            objective += elastic_objective(
                scores[view_index],
                target,
                self.regressors[view_index].coef_,
                self.alpha[view_index],
                self.l1_ratio[view_index],
            )
        return objective


# Define SCCA_IPLS class
class SCCA_IPLS(DeflationMixin, BaseIterative):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        accept_sparse=None,
        epochs=100,
        initialization: Union[str, callable] = "uniform",
        early_stopping=False,
        verbose=True,
        alpha=None,  # regularization parameter for Elastic
        l1_ratio=1,  # ratio of L1 to L2 penalty for Elastic
        positive=None,  # whether to enforce positive coefficients for Elastic
        stochastic=False,  # whether to use stochastic gradient descent for Elastic
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            tol=tol,
            accept_sparse=accept_sparse,
            epochs=epochs,
            initialization=initialization,
            early_stopping=early_stopping,
            verbose=verbose,
        )
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.positive = positive
        self.stochastic = stochastic

    def _check_params(self):
        self.alpha = _process_parameter("alpha", self.alpha, 0, self.n_views_)
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views_
        )
        self.l1_ratio = _process_parameter("l1_ratio", self.l1_ratio, 1, self.n_views_)
        self.regressors = initialize_regressors(
            self.alpha,
            self.l1_ratio,
            self.positive,
            self.stochastic,
            self.tol,
            self.random_state,
        )

    def _update_weights(self, views: Iterable[np.ndarray], i: int):
        # Update the weights for the current view using IPLS
        # Get the scores of all views
        scores = np.stack(self.transform(views))

        # Create a mask that is True for elements not equal to j along dim j
        mask = np.arange(scores.shape[0]) != i
        # Apply the mask to scores and sum along dim j
        target = np.sum(scores[mask], axis=0)
        # Solve the elastic regression
        self.regressors[i] = self.regressors[i].fit(views[i], target)
        # Normalize the coefficients by dividing by the square root of the covariance of the view and the coefficients
        self.regressors[i].coef_ /= np.sqrt(np.cov(views[i] @ self.regressors[i].coef_))
        # Update the weights with the coefficients of each regressor
        new_weights = self.regressors[i].coef_
        # Return the updated weights
        return new_weights[:, None]

    def _objective(self, views: Iterable[np.ndarray]):
        scores = np.stack(self.transform(views))
        objective = 0
        for view_index, view in enumerate(views):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            objective += elastic_objective(
                scores[view_index],
                target,
                self.regressors[view_index].coef_,
                self.alpha[view_index],
                self.l1_ratio[view_index],
            )
        return objective


def initialize_regressors(alpha, l1_ratio, positive, stochastic, tol, random_state):
    regressors = []
    for alpha, l1_ratio, positive in zip(alpha, l1_ratio, positive):
        if stochastic:
            regressors.append(
                SGDRegressor(
                    penalty="elasticnet",
                    alpha=alpha,
                    l1_ratio=l1_ratio,
                    fit_intercept=False,
                    tol=tol,
                    warm_start=True,
                    random_state=random_state,
                )
            )
        elif l1_ratio == 0:
            regressors.append(
                Ridge(
                    alpha=alpha,
                    fit_intercept=False,
                    positive=positive,
                    random_state=random_state,
                    tol=tol,
                )
            )
        elif l1_ratio == 1:
            regressors.append(
                Lasso(
                    alpha=alpha,
                    fit_intercept=False,
                    warm_start=True,
                    positive=positive,
                    random_state=random_state,
                    tol=tol,
                    selection="random",
                )
            )
        else:
            regressors.append(
                ElasticNet(
                    alpha=alpha,
                    l1_ratio=l1_ratio,
                    fit_intercept=False,
                    warm_start=True,
                    positive=positive,
                    random_state=random_state,
                    tol=tol,
                    selection="random",
                )
            )
    return regressors


def elastic_objective(z, y, w, alpha, l1_ratio):
    n = len(y)
    objective = np.linalg.norm(z - y) ** 2 / (2 * n)
    l1_pen = alpha * l1_ratio * np.linalg.norm(w, ord=1)
    l2_pen = alpha * (1 - l1_ratio) * np.linalg.norm(w, ord=2)
    return objective + l1_pen + l2_pen



# Import the necessary modules
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import numpy as np
import pandas as pd
import seaborn as sns

from cca_zoo.utils.check_values import check_seaborn_support


class ExplainedVarianceDisplay:
    """
    Display the explained variance of the latent variables of the views.

    Parameters
    ----------
    explained_variance_train : np.ndarray
        The explained variance of the train data.
    explained_variance_test : np.ndarray
        The explained variance of the test data.
    ratio : bool
        Whether to plot the ratio of explained variance or not.
    **kwargs : dict
        Keyword arguments to be passed to the seaborn lineplot.

    Attributes
    ----------
    figure_ : matplotlib.pyplot.figure
        The figure of the plot.

    Examples
    --------
    >>> from cca_zoo.visualisation import ExplainedVarianceDisplay
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from cca_zoo.linear import MCCA
    >>>
    >>> # Generate Sample Data
    >>> # --------------------
    >>> X = np.random.rand(100, 10)
    >>> Y = np.random.rand(100, 10)
    >>>
    >>> # Splitting the data into training and testing sets
    >>> X_train, X_test = X[:50], X[50:]
    >>> Y_train, Y_test = Y[:50], Y[50:]
    >>>
    >>> views = [X_train, Y_train]
    >>> test_views = [X_test, Y_test]
    >>>
    >>> # Train an MCCA Model
    >>> # -------------------
    >>> mcca = MCCA(latent_dimensions=2)
    >>> mcca.fit(views)
    >>>
    >>> # %%
    >>> # Plotting the Explained Variance
    >>> # ---------------------------------
    >>> ExplainedVarianceDisplay.from_estimator(mcca, views, test_views=test_views).plot()
    >>> plt.show()

    """

    def __init__(
        self,
        explained_variance_train,
        explained_variance_test=None,
        ratio=True,
        view_labels=None,
        **kwargs,
    ):
        self.explained_variance_train = explained_variance_train
        self.explained_variance_test = explained_variance_test
        self.ratio = ratio
        if view_labels is not None:
            assert len(view_labels) == len(
                self.explained_variance_train
            ), "view_labels must be the same length as train_views"
            assert len(view_labels) == len(
                self.explained_variance_test
            ), "view_labels must be the same length as test_views"
            self.view_labels = view_labels
        else:
            self.view_labels = [
                f"View {i}" for i in range(len(self.explained_variance_train))
            ]
        self.kwargs = kwargs

    def _validate_plot_params(self):
        check_seaborn_support("CorrelationHeatmapDisplay")

    @classmethod
    def from_estimator(
        cls, model, train_views, test_views=None, ratio=True, view_labels=None, **kwargs
    ):
        # explained_variance_train will be a numpy array of shape (latent_dimensions,len(train_views))
        if ratio:
            explained_variance_train = model.explained_variance_ratio(train_views)
        else:
            explained_variance_train = model.explained_variance(train_views)
        if test_views is not None:
            if ratio:
                explained_variance_test = model.explained_variance_ratio(test_views)
            else:
                explained_variance_test = model.explained_variance(test_views)
        else:
            explained_variance_test = None
        if ratio:
            return cls.from_explained_variance_ratio(
                explained_variance_train,
                explained_variance_test,
                view_labels=view_labels,
                **kwargs,
            )
        else:
            return cls.from_explained_variance(
                explained_variance_train,
                explained_variance_test,
                view_labels=view_labels,
                **kwargs,
            )

    @classmethod
    def from_explained_variance(
        cls,
        explained_variance_train,
        explained_variance_test=None,
        view_labels=None,
        **kwargs,
    ):
        return cls(
            explained_variance_train,
            explained_variance_test,
            ratio=False,
            view_labels=view_labels,
            **kwargs,
        )

    @classmethod
    def from_explained_variance_ratio(
        cls,
        explained_variance_train,
        explained_variance_test=None,
        view_labels=None,
        **kwargs,
    ):
        return cls(
            explained_variance_train,
            explained_variance_test,
            ratio=True,
            view_labels=view_labels,
            **kwargs,
        )

    def plot(self, ax=None):
        self._validate_plot_params()
        # Use seaborn lineplot with style='Train' and hue='View' to plot the train and test data
        # Reshape the data so that each row has a 'value', 'view index', and 'train' column
        data = pd.DataFrame(self.explained_variance_train, index=self.view_labels).T
        # Give the index a name so that it can be used as a column later
        data.index.name = "Latent dimension"
        # Melt the dataframe so that each row has a 'value', 'view index', and 'train' column
        data = data.reset_index().melt(
            id_vars="Latent dimension", value_vars=self.view_labels
        )
        data.columns = ["Latent dimension", "View", "value"]
        data["Mode"] = "Train"  # Add a column indicating train data
        if self.explained_variance_test is not None:
            data_test = pd.DataFrame(
                self.explained_variance_test, index=self.view_labels
            ).T
            # Give the index a name so that it can be used as a column later
            data_test.index.name = "Latent dimension"
            # Melt the dataframe so that each row has a 'value', 'view index', and 'train' column
            data_test = data_test.reset_index().melt(
                id_vars="Latent dimension", value_vars=self.view_labels
            )
            data_test.columns = ["Latent dimension", "View", "value"]
            data_test["Mode"] = "Test"  # Add a column indicating train data
            data = pd.concat([data, data_test])  # Concatenate the two dataframes
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 5))
        else:
            fig = ax.get_figure()
        sns.lineplot(
            data=data,
            x="Latent dimension",
            y="value",
            hue="View",
            style="Mode",
            marker="o",
            ax=ax,
        )
        # Set x-ticks to integers
        ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True))
        ax.set_xlabel("Latent dimension")
        if self.ratio:
            ax.set_ylabel("Explained Variance %")
            ax.yaxis.set_major_formatter(mtick.PercentFormatter(1.0))
        else:
            ax.set_ylabel("Explained Variance")
        ax.set_title("Explained Variance")
        plt.tight_layout()
        self.figure_ = fig
        return self



import torch
from pytorch_lightning import Callback, LightningModule, Trainer

from cca_zoo.deep.objectives import MCCA


class BatchValidationCorrelationCallback(Callback):
    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        val_corr = pl_module.score(trainer.val_dataloaders).sum()
        pl_module.log(
            "val/corr",
            val_corr,
        )


class BatchTrainCorrelationCallback(Callback):
    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        train_corr = pl_module.score(trainer.train_dataloader).sum()
        pl_module.log(
            "train/corr",
            train_corr,
        )


class MinibatchTrainCorrelationCallback(Callback):
    mcca = MCCA()

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs,
        batch,
        batch_idx,
        dataloader_idx,
    ):
        with torch.no_grad():
            train_corr = self.mcca.loss(pl_module(batch["views"])).sum()
            pl_module.log(
                "train/corr",
                train_corr,
            )


class MinibatchValidationCorrelationCallback(Callback):
    mcca = MCCA()

    def on_validation_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs,
        batch,
        batch_idx,
        dataloader_idx,
    ):
        with torch.no_grad():
            val_corr = self.mcca.loss(pl_module(batch["views"])).sum()
            pl_module.log(
                "val/corr",
                val_corr,
            )



import itertools
import warnings
from typing import Union, Iterable

import numpy as np

from cca_zoo.linear._iterative._base import BaseIterative
from cca_zoo.linear._iterative._deflation import DeflationMixin
from cca_zoo.linear._search import _delta_search
from cca_zoo.utils import _process_parameter


class SCCA_PMD(DeflationMixin, BaseIterative):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        accept_sparse=None,
        epochs=100,
        initialization: Union[str, callable] = "pls",
        early_stopping=False,
        verbose=True,
        tau=None,  # regularization parameter for PMD
        positive=False,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            tol=tol,
            accept_sparse=accept_sparse,
            epochs=epochs,
            initialization=initialization,
            early_stopping=early_stopping,
            verbose=verbose,
        )
        self.tau = tau
        self.positive = positive

    def _check_params(self):
        if self.tau is None:
            warnings.warn(
                "tau parameter not set. Setting to tau=1 i.e. maximum regularisation of l1 norm"
            )
        self.tau = _process_parameter("tau", self.tau, 1, self.n_views_)
        if any(tau < 0 or tau > 1 for tau in self.tau):
            raise ValueError(
                "All regularisation parameters should be between 0 and 1 "
                f"1. tau=[{self.tau}]"
            )
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views_
        )

    def _update_weights(self, views: np.ndarray, i: int):
        if not hasattr(self, "t"):
            shape_sqrts = [np.sqrt(weight.shape[0]) for weight in self.weights]
            self.t = [max(1, x * y) for x, y in zip(self.tau, shape_sqrts)]
        # Update the weights for the current view using PMD
        # Get the scores of all views
        scores = np.stack(self.transform(views))
        # Create a mask that is True for elements not equal to i along dim i
        mask = np.arange(scores.shape[0]) != i
        # Apply the mask to scores and sum along dim i
        target = np.sum(scores[mask], axis=0)
        # Compute the new weights by multiplying the view with the target
        new_weights = views[i].T @ target
        if self.positive[i]:
            # If positive is true, set all negative values to 0
            new_weights[new_weights < 0] = 0
        # Apply the delta search function to the new weights with the regularization parameter
        new_weights = _delta_search(new_weights, self.t[i], tol=self.tol)
        # Return the new weights
        return new_weights

    def _objective(self, views: Iterable[np.ndarray]):
        # Compute the objective function value for a given set of views using SCCA
        # Get the scores of all views
        transformed_views = self.transform(views)
        all_covs = []
        # Sum all the pairwise covariances except self covariance
        for x, y in itertools.product(transformed_views, repeat=2):
            all_covs.append(
                np.diag(
                    np.corrcoef(x.T, y.T)[
                        : self.latent_dimensions, self.latent_dimensions :
                    ]
                )
            )
        # the sum of covariances
        return np.sum(all_covs) - np.sum(
            [
                self.tau[i] * np.linalg.norm(self.weights[i])
                for i in range(len(self.weights))
            ]
        )



from typing import Iterable

import numpy as np
from sklearn.utils.validation import check_is_fitted

from cca_zoo.linear._mcca import MCCA


class PartialCCA(MCCA):
    r"""
    A class used to fit a partial CCA model. This model extends CCA to account for confounding variables that may affect the correlation between views.

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        w_i^TX_i^TX_iw_i=1

        w_i^TX_i^TZ=0


    Example
    -------
    >>> from cca_zoo.linear import PartialCCA
    >>> X1 = np.random.rand(10,5)
    >>> X2 = np.random.rand(10,5)
    >>> partials = np.random.rand(10,3)
    >>> model = PartialCCA()
    >>> model.fit((X1,X2),partials=partials).score((X1,X2))

    References
    ----------
    Rao, B. Raja. "Partial canonical correlations." Trabajos de estadistica y de investigación operativa 20.2-3 (1969): 211-219.

    """

    def fit(self, views: Iterable[np.ndarray], y=None, partials=None, **kwargs):
        self.pca = False
        super().fit(
            views, y=y, partials=partials, **kwargs
        )  # call the parent class fit method

    def _process_data(self, views, partials=None, **kwargs):
        if partials is None:
            raise ValueError(
                f"partials is {partials}. Require matching partials to transform with"
                f"partial CCA."
            )
        self.confound_betas = [
            np.linalg.pinv(partials) @ view for view in views
        ]  # compute the confounding betas for each view using pseudo-inverse of partials
        views = [
            view
            - partials
            @ np.linalg.pinv(partials)
            @ view  # remove the confounding effect from each view using projection matrix
            for view, confound_beta in zip(views, self.confound_betas)
        ]
        return views

    def transform(self, views: Iterable[np.ndarray], partials=None, **kwargs):
        if partials is None:
            raise ValueError(
                f"partials is {partials}. Require matching partials to transform with"
                f"partial CCA."
            )
        check_is_fitted(
            self, attributes=["weights"]
        )  # check if the model has been fitted before transforming
        transformed_views = []
        for i, (view) in enumerate(views):
            transformed_view = (
                view
                - partials
                @ self.confound_betas[
                    i
                ]  # remove the confounding effect from each view using stored confounding betas
            ) @ self.weights[
                i
            ]  # multiply each view by its corresponding weight matrix
            transformed_views.append(
                transformed_view
            )  # append the transformed view to the list of transformed views
        return transformed_views  # return the list of transformed views

    def _more_tags(self):
        return {"multiview": True}  # indicate that this model can handle multiview data



import itertools
from abc import abstractmethod
from typing import Iterable, Union, List, Optional, Any

import numpy as np
from numpy.linalg import svd
from scipy.linalg import block_diag
from sklearn.base import BaseEstimator, MultiOutputMixin, RegressorMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted, check_array


class BaseModel(BaseEstimator, MultiOutputMixin, RegressorMixin):
    """
    A base class for multivariate latent variable linear.

    This class implements common methods and attributes for fitting and transforming
    multiple views of data using latent variable linear. It inherits from scikit-learn's
    BaseEstimator, MultiOutputMixin and RegressorMixin classes.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to fit. Default is 1.
    copy_data : bool, optional
        Whether to copy the data. Default is True.
    accept_sparse : bool, optional
        Whether to accept sparse data. Default is False.
    random_state : int, RandomState instance or None, optional (default=None)
        Pass an int for reproducible output across multiple function calls.

    Attributes
    ----------
    n_views_ : int
        Number of views.
    n_features_ : list of int
        Number of features for each view.

    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        accept_sparse=False,
        random_state: Union[int, np.random.RandomState] = None,
    ):
        self.latent_dimensions = latent_dimensions
        self.copy_data = copy_data
        self.accept_sparse = accept_sparse
        self.random_state = check_random_state(random_state)
        self.dtypes = FLOAT_DTYPES
        self._loadings = None

    def _validate_data(self, views: Iterable[np.ndarray]):
        if not all(view.shape[0] == views[0].shape[0] for view in views):
            raise ValueError("All views must have the same number of samples")
        if not all(view.ndim == 2 for view in views):
            raise ValueError("All views must have 2 dimensions")
        if not all(view.dtype in self.dtypes for view in views):
            raise ValueError("All views must have dtype of {}.".format(self.dtypes))
        if not all(view.shape[1] >= self.latent_dimensions for view in views):
            raise ValueError(
                "All views must have at least {} features.".format(
                    self.latent_dimensions
                )
            )
        self.n_views_ = len(views)
        self.n_features_ = [view.shape[1] for view in views]
        self.n_samples_ = views[0].shape[0]
        if self.copy_data:
            return [check_array(view, copy=True, dtype=self.dtypes) for view in views]
        else:
            return [check_array(view, copy=False, dtype=self.dtypes) for view in views]

    def _check_params(self):
        """
        Checks the parameters of the model.
        """
        pass

    @abstractmethod
    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        """
        Fits the model to the given data

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        y : None
        kwargs: any additional keyword arguments required by the given model

        Returns
        -------
        self : object

        """
        return self

    def transform(self, views: Iterable[np.ndarray], **kwargs) -> List[np.ndarray]:
        """

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        transformed_views : list of numpy arrays

        """
        check_is_fitted(self, attributes=["weights"])
        transformed_views = []
        for i, view in enumerate(views):
            transformed_view = view @ self.weights[i]
            transformed_views.append(transformed_view)
        return transformed_views

    def fit_transform(self, views: Iterable[np.ndarray], **kwargs) -> List[np.ndarray]:
        """
        Fits the model to the given data and returns the transformed views

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        transformed_views : list of numpy arrays

        """
        return self.fit(views, **kwargs).transform(views, **kwargs)

    def pairwise_correlations(
        self, views: Iterable[np.ndarray], **kwargs
    ) -> np.ndarray:
        """
        Returns the pairwise correlations between the views in each dimension

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        pairwise_correlations : numpy array of shape (n_views, n_views, latent_dimensions)

        """
        transformed_views = self.transform(views, **kwargs)
        all_corrs = []
        for x, y in itertools.product(transformed_views, repeat=2):
            all_corrs.append(
                np.diag(
                    np.corrcoef(x.T, y.T)[
                        : self.latent_dimensions, self.latent_dimensions :
                    ]
                )
            )
        all_corrs = np.array(all_corrs).reshape(
            (self.n_views_, self.n_views_, self.latent_dimensions)
        )
        return all_corrs

    def score(
        self, views: Iterable[np.ndarray], y: Optional[Any] = None, **kwargs
    ) -> float:
        """
        Returns the average pairwise correlation between the views

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        y : None
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        score : float


        """
        # by default return the average pairwise correlation in each dimension (for 2 views just the correlation)
        pair_corrs = self.pairwise_correlations(views, **kwargs)
        # sum all the pairwise correlations for each dimension. Subtract the self correlations. Divide by the number of views. Gives average correlation
        dim_corrs = np.sum(pair_corrs, axis=(0, 1)) - pair_corrs.shape[0]
        # number of pairs is n_views choose 2
        num_pairs = (self.n_views_ * (self.n_views_ - 1)) / 2
        dim_corrs = dim_corrs / (2 * num_pairs)
        return dim_corrs

    def canonical_loadings(
        self, views: Iterable[np.ndarray], normalize: bool = True, **kwargs
    ) -> List[np.ndarray]:
        """
        Calculate canonical loadings for each view.

        Canonical loadings represent the correlation between the original variables
        in a view and their respective canonical variates. Canonical variates are
        linear combinations of the original variables formed to maximize the correlation
        with canonical variates from another view.

        Mathematically, given two views \(X_i\), canonical variates
        from the views are:

            \(Z_i = w_i^T X_i\)

        The canonical loading for a variable in \(X_i\) is the correlation between
        that variable and \(Z_i\).

        Parameters
        ----------
        views : list/tuple of numpy arrays
            Each array corresponds to a view. All views must have the same number of rows (observations).

        Returns
        -------
        loadings : list of numpy arrays
            Canonical loadings for each view. High absolute values indicate that
            the respective original variables play a significant role in defining the canonical variate.

        """
        transformed_views = self.transform(views, **kwargs)
        if normalize:
            loadings = [
                np.corrcoef(view, transformed_view, rowvar=False)[
                    : view.shape[1], view.shape[1] :
                ]
                for view, transformed_view in zip(views, transformed_views)
            ]
        else:
            loadings = [
                view.T @ transformed_view
                for view, transformed_view in zip(views, transformed_views)
            ]
        return loadings

    @property
    def loadings(self) -> List[np.ndarray]:
        """
        Compute and return loadings for each view. These are cached for performance optimization.

        In the context of the cca-zoo models, loadings are the normalized weights. Due to the structure of these models,
        weight vectors are normalized such that w'X'Xw = 1, as opposed to w'w = 1, which is commonly used in PCA.
        As a result, when computing the loadings, the weights are normalized to have unit norm, ensuring that the loadings
        range between -1 and 1.

        It's essential to differentiate between these loadings and canonical loadings. The latter are correlations between
        the original variables and their corresponding canonical variates.

        Returns
        -------
        List[np.ndarray]
            Loadings for each view.
        """
        check_is_fitted(self, attributes=["weights"])
        if self._loadings is None:
            # Compute loadings only if they haven't been computed yet
            self._loadings = [
                weights / np.linalg.norm(weights, axis=0) for weights in self.weights
            ]
        return self._loadings

    def explained_variance(self, views: Iterable[np.ndarray]) -> List[np.ndarray]:
        """
        Calculates the variance captured by each latent dimension for each view.

        Returns
        -------
        transformed_vars : list of numpy arrays
        """
        check_is_fitted(self, attributes=["weights"])

        # Transform the views using the loadings
        transformed_views = [
            view @ loading for view, loading in zip(views, self.loadings)
        ]

        # Calculate the variance of each latent dimension in the transformed views
        transformed_vars = [
            np.var(transformed, axis=0) for transformed in transformed_views
        ]

        return transformed_vars

    def explained_variance_ratio(self, views: Iterable[np.ndarray]) -> List[np.ndarray]:
        """
        Calculates the ratio of the variance captured by each latent dimension to the total variance for each view.

        Returns
        -------
        explained_variance_ratios : list of numpy arrays
        """
        total_vars = [
            (np.sum(s**2) / (view.shape[0] - 1))
            for view in views
            for _, s, _ in [svd(view)]
        ]

        transformed_vars = self.explained_variance(views)

        # Calculate the explained variance ratio for each latent dimension for each view
        explained_variance_ratios = [
            transformed_var / total_var
            for transformed_var, total_var in zip(transformed_vars, total_vars)
        ]

        return explained_variance_ratios

    def explained_variance_cumulative(
        self, views: Iterable[np.ndarray]
    ) -> List[np.ndarray]:
        """
        Calculates the cumulative explained variance ratio for each latent dimension for each view.

        Returns
        -------
        cumulative_ratios : list of numpy arrays
        """
        ratios = self.explained_variance_ratio(views)
        cumulative_ratios = [np.cumsum(ratio) for ratio in ratios]

        return cumulative_ratios

    def _compute_covariance(self, views: Iterable[np.ndarray]) -> np.ndarray:
        """
        Computes the covariance matrix for the given views.

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)

        Returns
        -------
        cov : numpy array
            Computed covariance matrix.
        """
        cov = np.cov(np.hstack(views), rowvar=False)
        cov -= block_diag(*[np.cov(view, rowvar=False) for view in views])
        return cov

    def explained_covariance(self, views: Iterable[np.ndarray]) -> np.ndarray:
        """
        Calculates the covariance matrix of the transformed components for each view.

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)

        Returns
        -------
        explained_covariances : list of numpy arrays
            Covariance matrices for the transformed components of each view.
        """
        check_is_fitted(self, attributes=["weights"])

        # Transform the views using the loadings
        transformed_views = [
            view @ loading for view, loading in zip(views, self.loadings)
        ]

        k = transformed_views[0].shape[1]

        explained_covariances = np.zeros(k)

        # just take the kth column of each transformed view and _compute_covariance
        for i in range(k):
            transformed_views_k = [view[:, i][:, None] for view in transformed_views]
            cov_ = self._compute_covariance(transformed_views_k)
            _, s_, _ = svd(cov_)
            explained_covariances[i] = s_[0]

        return explained_covariances

    def explained_covariance_ratio(self, views: Iterable[np.ndarray]) -> np.ndarray:
        minimum_dimension = min([view.shape[1] for view in views])

        cov = self._compute_covariance(views)
        _, S, _ = svd(cov)
        # select every other element starting from the first until the minimum dimension
        total_explained_covariance = S[::2][:minimum_dimension].sum()

        explained_covariances = self.explained_covariance(views)
        explained_covariance_ratios = explained_covariances / total_explained_covariance

        return explained_covariance_ratios

    def explained_covariance_cumulative(
        self, views: Iterable[np.ndarray]
    ) -> np.ndarray:
        """
        Calculates the cumulative explained covariance ratio for each latent dimension for each view.

        Returns
        -------
        cumulative_ratios : list of numpy arrays
        """
        ratios = self.explained_covariance_ratio(views)
        cumulative_ratios = [np.cumsum(ratio) for ratio in ratios]

        return cumulative_ratios

    def predict(self, views: Iterable[np.ndarray]) -> List[np.ndarray]:
        """
        Predicts the missing view from the given views.


        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)

        Returns
        -------
        predicted_views : list of numpy arrays. None if the view is missing.
            Predicted views.

        Examples
        --------
        >>> import numpy as np
        >>> X1 = np.random.rand(100, 5)
        >>> X2 = np.random.rand(100, 5)
        >>> cca = CCA()
        >>> cca.fit([X1, X2])
        >>> X1_pred, X2_pred = cca.predict([X1, None])

        """
        check_is_fitted(self, attributes=["weights"])
        # check if views is same length as weights
        if len(views) != len(self.weights):
            raise ValueError(
                "The number of views must be the same as the number of weights. Put None for missing views."
            )
        transformed_views = []
        for i, view in enumerate(views):
            if view is not None:
                transformed_view = view @ self.weights[i]
                transformed_views.append(transformed_view)
        # average the transformed views
        average_score = np.mean(transformed_views, axis=0)
        # return the average score transformed back to the original space
        reconstucted_views = []
        for i, view in enumerate(views):
            reconstructed_view = average_score @ np.linalg.pinv(self.weights[i])
            reconstucted_views.append(reconstructed_view)
        return reconstucted_views



# from itertools import combinations
# from typing import Iterable, Union
#
# import numpy as np
#
# from cca_zoo.linear._iterative._base import BaseIterative
# from cca_zoo.linear._iterative._deflation import DeflationMixin
# from cca_zoo.linear._search import _delta_search, support_threshold
# from cca_zoo.utils import _process_parameter
#
#
# class SWCCA(BaseIterative, DeflationMixin):
#     r"""
#     A class used to fit SWCCA model
#
#     References
#     ----------
#     .. Wenwen, M. I. N., L. I. U. Juan, and Shihua Zhang. "Sparse Weighted Canonical Correlation Analysis." Chinese Journal of Electronics 27.3 (2018): 459-466.
#
#     Examples
#     --------
#     >>> from cca_zoo.linear import SWCCA
#     >>> import numpy as np
#     >>> rng=np.random.RandomState(0)
#     >>> X1 = rng.random((10,5))
#     >>> X2 = rng.random((10,5))
#     >>> model = SWCCA(regularisation='l0',tau=[2, 2], sample_support=5, random_state=0)
#     >>> model.fit((X1,X2)).score((X1,X2))
#     array([0.61620969])
#     """
#
#     def __init__(
#         self,
#         copy_data=True,
#         random_state=None,
#         max_iter: int = 500,
#         initialization: str = "random",
#         tol: float = 1e-3,
#         regularisation="l0",
#         tau: Union[Iterable[Union[float, int]], Union[float, int]] = None,
#         sample_support=None,
#         positive=False,
#         verbose=0,
#     ):
#         self.tau = tau
#         self.sample_support = sample_support
#         if regularisation == "l0":
#             self.update = support_threshold
#         elif regularisation == "l1":
#             self.update = _delta_search
#         self.regularisation = regularisation
#         self.positive = positive
#         super().__init__(
#             latent_dims=1,
#             copy_data=copy_data,
#             max_iter=max_iter,
#             initialization=initialization,
#             tol=tol,
#             random_state=random_state,
#             verbose=verbose,
#         )
#
#     def _check_params(self):
#         if self.sample_support is None:
#             self.sample_support = self.n
#         self.tau = _process_parameter("tau", self.tau, 2, self.n_views)
#         self.positive = _process_parameter(
#             "positive", self.positive, False, self.n_views
#         )
#
#     def _initialize(self, views):
#         self.sample_weights = np.ones(self.n)
#         self.sample_weights /= np.linalg.norm(self.sample_weights)
#
#     def _update(self, views, scores, weights):
#         # Update each view using loop update function
#         for view_index, view in enumerate(views):
#             targets = np.ma.array(scores, mask=False)
#             targets.mask[view_index] = True
#             weights[view_index] = (
#                 views[view_index] * self.sample_weights[:, np.newaxis]
#             ).T @ targets.sum(axis=0).filled()
#             weights[view_index] = self.update(
#                 weights[view_index],
#                 self.tau[view_index],
#                 positive=self.positive[view_index],
#             )
#             weights[view_index] /= np.linalg.norm(weights[view_index])
#             if view_index == self.n_views - 1:
#                 self.sample_weights = self._update_sample_weights(scores)
#             scores[view_index] = views[view_index] @ weights[view_index]
#         return scores, weights
#
#     def _update_sample_weights(self, scores):
#         w = scores.prod(axis=0)
#         sample_weights = support_threshold(w, self.sample_support)
#         sample_weights /= np.linalg.norm(sample_weights)
#         return sample_weights
#
#     def _objective(self, views, scores, weights) -> int:
#         # default objective is correlation
#         obj = 0
#         for score_i, score_j in combinations(scores, 2):
#             obj += (score_i * self.sample_weights).T @ score_j
#         return obj



import numpy as np

from cca_zoo.linear._mcca import MCCA, rCCA


def reduce_dims(x):
    U, S, _ = np.linalg.svd(x, full_matrices=False)
    return U @ np.diag(S)


class PLSMixin:
    def _more_tags(self):
        # Indicate that this class is for multiview data
        return {"pls": True}


class PLS(rCCA, PLSMixin):
    r"""
    A class used to fit a simple PLS model. This model finds the linear projections of two views that maximize their covariance.

    Implements PLS by inheriting regularised CCA with maximal regularisation. This is equivalent to solving the following optimization problem:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        w_1^Tw_1=1

        w_2^Tw_2=1

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state, by default None

    Examples
    --------
    >>> import numpy as np
    >>> X1 = np.random.rand(100, 5)
    >>> X2 = np.random.rand(100, 5)
    >>> pls = PLS(latent_dimensions=2)
    >>> pls.fit([X1, X2])
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
    ):
        # Call the parent class constructor with c=1 to enable maximal regularization
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            c=1,
            random_state=random_state,
        )


class MPLS(MCCA, PLSMixin):
    r"""
    A class used to fit a mutiview PLS model. This model finds the linear projections of two views that maximize their covariance.

    Implements PLS by inheriting regularised CCA with maximal regularisation. This is equivalent to solving the following optimization problem:

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state, by default None
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
    ):
        # Call the parent class constructor with c=1 to enable maximal regularization
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            c=1,
            random_state=random_state,
        )



from typing import Dict, Any

import torch

from ._dcca import DCCA


class DCCA_EY(DCCA):
    """

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A GeneralizedDeflation EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(self, latent_dimensions: int, encoders=None, r: float = 0, **kwargs):
        super().__init__(
            latent_dimensions=latent_dimensions, encoders=encoders, **kwargs
        )
        self.r = r

    def loss(self, batch, **kwargs):
        # Encoding the views with the forward method
        z = self(batch["views"])
        independent_views = batch.get("independent_views", None)
        # Getting A and B matrices from z
        A, B = self.get_AB(z)
        rewards = torch.trace(2 * A)
        if independent_views is None:
            penalties = torch.trace(B @ B)
        else:
            # Encoding another set of views with the forward method
            independent_z = self(independent_views)
            # Getting A' and B' matrices from independent_z
            independent_A, independent_B = self.get_AB(independent_z)
            penalties = torch.trace(B @ independent_B)
        return {
            "objective": -rewards + penalties,
            "rewards": rewards,
            "penalties": penalties,
        }

    def get_AB(self, z):
        A = torch.zeros(
            self.latent_dimensions, self.latent_dimensions, device=z[0].device
        )  # initialize the cross-covariance matrix
        B = torch.zeros(
            self.latent_dimensions, self.latent_dimensions, device=z[0].device
        )  # initialize the auto-covariance matrix
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    B += torch.cov(zi.T)  # add the auto-covariance of each view to B
                else:
                    A += torch.cov(torch.hstack((zi, zj)).T)[
                        self.latent_dimensions :, : self.latent_dimensions
                    ]  # add the cross-covariance of each pair of views to A
        return A / len(z), B / len(
            z
        )  # return the normalized matrices (divided by the number of views)



from typing import Iterable, List, Union

import numpy as np
import pytorch_lightning as pl
import torch
from torch.utils import data
from torch.utils.data import DataLoader

from cca_zoo._base import BaseModel
from cca_zoo.data.deep import NumpyDataset
from cca_zoo.linear._iterative._base import _default_initializer

# Default Trainer kwargs
DEFAULT_TRAINER_KWARGS = dict(
    enable_checkpointing=False,
    logger=False,
    enable_model_summary=False,
    enable_progress_bar=True,
)

DEFAULT_LOADER_KWARGS = dict(pin_memory=False, drop_last=False, shuffle=False)

DEFAULT_OPTIMIZER_KWARGS = dict(optimizer="Adam")


class BaseGradientModel(BaseModel, pl.LightningModule):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        accept_sparse=None,
        batch_size=None,
        dataloader_kwargs=None,
        epochs=1,
        learning_rate=1,
        initialization: Union[str, callable] = "random",
        trainer_kwargs=None,
        optimizer_kwargs=None,
    ):
        BaseModel.__init__(
            self,
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            accept_sparse=accept_sparse,
        )
        pl.LightningModule.__init__(self)
        self.tol = tol
        self.batch_size = batch_size
        self.epochs = epochs
        self.learning_rate = learning_rate
        # validate the initialization method
        if initialization not in ["random", "uniform", "unregularized", "pls"]:
            raise ValueError(
                "Initialization method must be one of ['random', 'uniform', 'unregularized', 'pls']"
            )
        else:
            self.initialization = initialization
        self.dataloader_kwargs = dataloader_kwargs or DEFAULT_LOADER_KWARGS
        self.trainer_kwargs = trainer_kwargs or DEFAULT_TRAINER_KWARGS
        self.optimizer_kwargs = optimizer_kwargs or DEFAULT_OPTIMIZER_KWARGS

    def fit(self, views: Iterable[np.ndarray], y=None, validation_views=None, **kwargs):
        views = self._validate_data(views)
        if validation_views is not None:
            validation_views = self._validate_data(validation_views)
        self._check_params()
        self._initialize(views)
        self.weights = self._fit(views, validation_views=validation_views)
        return self

    def _fit(self, views: Iterable[np.ndarray], validation_views=None):
        # Set the weights attribute as torch parameters with gradients
        self.torch_weights = [
            torch.nn.Parameter(torch.from_numpy(weight), requires_grad=True)
            for weight in self.weights
        ]
        self.torch_weights = torch.nn.ParameterList(self.torch_weights)
        # make a trainer
        trainer = pl.Trainer(
            max_epochs=self.epochs,
            **self.trainer_kwargs,
        )
        train_dataset, val_dataset = self.get_dataset(
            views, validation_views=validation_views
        )
        train_dataloader, val_dataloader = self.get_dataloader(
            train_dataset, val_dataset
        )
        if self.batch_size is None:
            # if the batch size is None, put views on the device
            self.batch = {
                "views": [
                    view.to(trainer._accelerator_connector._accelerator_flag)
                    for view in train_dataset.views
                ]
            }
        trainer.fit(self, train_dataloader, val_dataloader)
        # return the weights from the module. They will need to be changed from torch tensors to numpy arrays
        weights = [weight.detach().cpu().numpy() for weight in self.torch_weights]
        return weights

    def get_dataset(self, views: Iterable[np.ndarray], validation_views=None):
        dataset = NumpyDataset(views) if self.batch_size else FullBatchDataset(views)
        if validation_views is not None:
            val_dataset = (
                NumpyDataset(validation_views)
                if self.batch_size
                else FullBatchDataset(validation_views)
            )
        else:
            val_dataset = None
        return dataset, val_dataset

    def get_dataloader(self, train_dataset, val_dataset):
        train_loader = DataLoader(
            train_dataset,
            batch_size=self.batch_size,
            **self.dataloader_kwargs,
        )
        if val_dataset is not None:
            val_loader = DataLoader(
                val_dataset,
                batch_size=self.batch_size,
                **self.dataloader_kwargs,
            )
        else:
            val_loader = None
        return train_loader, val_loader

    def _initialize(self, views: Iterable[np.ndarray]):
        """Initialize the CCA weights using the initialization method or function.

        Parameters
        ----------
        views : Iterable[np.ndarray]
            The input views to initialize the CCA weights from
        """
        pls = self._get_tags().get("pls", False)
        initializer = _default_initializer(
            self.initialization, self.random_state, self.latent_dimensions, pls
        )
        # Fit the initializer on the input views and get the weights as numpy arrays
        self.weights = initializer.fit(views).weights
        self.weights = [weights.astype(np.float32) for weights in self.weights]

    def _more_tags(self):
        # Indicate that this class is for multiview data
        return {"iterative": True}

    def forward(self, views: List[torch.Tensor]) -> List[torch.Tensor]:
        """Perform a forward pass on the input views.

        Args:
            views (List[torch.Tensor]): The input views as torch tensors.

        Returns:
            List[torch.Tensor]: The output views as torch tensors.
        """
        return [view @ weight for view, weight in zip(views, self.torch_weights)]

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """Configure the optimizer for the loop.

        Returns:
            torch.optim.Optimizer: The optimizer object.
        """
        # construct optimizer using optimizer_kwargs
        optimizer_name = self.optimizer_kwargs.get("optimizer", "SGD")
        kwargs = self.optimizer_kwargs.copy()
        kwargs.pop("optimizer", None)
        optimizer = getattr(torch.optim, optimizer_name)(
            self.torch_weights, lr=self.learning_rate, **kwargs
        )
        return optimizer

    def objective(self, *args, **kwargs) -> float:
        """Compute the objective value for the loop.

        This method should be implemented by subclasses.

        Returns:
            float: The objective value.

        Raises:
            NotImplementedError: If the method is not implemented by subclasses.
        """
        raise NotImplementedError


class FullBatchDataset(data.Dataset):
    def __init__(self, views: Iterable[np.ndarray]):
        self.views = [torch.from_numpy(view).float() for view in views]

    def __len__(self):
        return 1

    def __getitem__(self, index):
        return index



import torch

from .._base import BaseDeep
from .._generative._base import _GenerativeMixin
from ..architectures import Encoder


class SplitAE(BaseDeep, _GenerativeMixin):
    """
    A class used to fit a Split Autoencoder model.

    References
    ----------
    Ngiam, Jiquan, et al. "Multimodal deep learning." ICML. 2011.

    """

    def __init__(
        self,
        latent_dimensions: int,
        encoder=Encoder,
        decoders=None,
        latent_dropout=0,
        recon_loss_type="mse",
        img_dim=None,
        **kwargs
    ):
        """

        :param latent_dimensions: # latent dimensions
        :param encoder: list of encoder networks
        :param decoders:  list of decoder networks
        """
        super().__init__(latent_dimensions=latent_dimensions, **kwargs)
        self.img_dim = img_dim
        self.encoder = encoder
        self.decoders = torch.nn.ModuleList(decoders)
        self.latent_dropout = torch.nn.Dropout(p=latent_dropout)
        self.recon_loss_type = recon_loss_type

    def forward(self, views, **kwargs):
        """
        Forward method for the model. Outputs latent encoding for each view

        :param views:
        :param kwargs:
        :return:
        """
        z = []
        z.append(self.encoder(views[0]))
        return z

    def _decode(self, z, **kwargs):
        """
        This method is used to decode from the latent space to the best prediction of the original views

        :param z:
        """
        recon = []
        for i, decoder in enumerate(self.decoders):
            recon.append(decoder(self.latent_dropout(z[0])))
        return recon

    def loss(self, batch, **kwargs):
        z = self(batch["views"])
        recons = self._decode(z)
        loss = dict()
        loss["reconstruction"] = torch.stack(
            [
                self.recon_loss(x, recon, loss_type=self.recon_loss_type)
                for x, recon in zip(batch["views"], recons)
            ]
        ).sum()
        loss["objective"] = loss["reconstruction"]
        return loss



import numpy as np
from scipy.optimize import minimize


def _delta_search(w, c, tol=1e-8):
    """
    Searches for threshold delta such that the 1-norm of weights w is less than or equal to c and the 2-norm is equal to 1.
    Parameters
    ----------
    w : numpy array
        weights found by one power method iteration
    c : float
        1-norm threshold
    init : float, optional
        initial value for delta (default is 0)

    Returns
    -------
    numpy array
        updated weights

    """
    # First normalize the weights to unit length
    w = w / np.linalg.norm(w)

    # Define a scalar function that returns the difference between the 1-norm of coefficients and the threshold c
    def f(delta):
        # Apply soft thresholding to the weights with delta
        coef = np.clip(w - delta, 0, None) - np.clip(-w - delta, 0, None)

        if np.sum(coef**2) == 0:
            coef[:] = 1000
        else:
            # Normalize the coefficients to unit length if nonzero
            coef /= np.linalg.norm(coef)

        # Return the square of the difference between the 1-norm of coefficients and the threshold c
        return (np.sum(np.abs(coef)) - c) ** 2

    # Find the minimum of f using scipy minimization function
    # You can specify the method or let the function choose the best one for you
    # You can also pass other parameters like tol, maxiter, etc.
    # bound x to be between 0 and 1

    # try some different methods until one gets result.success == True
    result = minimize(f, x0=0, bounds=[(0, 1)], method="L-BFGS-B", tol=tol)
    if not result.success:
        result = minimize(f, x0=0, bounds=[(0, 1)], method="TNC", tol=tol)
    if not result.success:
        result = minimize(f, x0=0, bounds=[(0, 1)], method="SLSQP", tol=tol)
    if not result.success:
        result = minimize(f, x0=0, bounds=[(0, 1)], method="trust-constr", tol=tol)

    # Check if the solution is valid and converged
    if result.success:
        # Get the optimal delta from the result object
        delta = result.x

        # Apply soft thresholding to the weights with optimal delta
        coef = np.where(w - delta > 0, w - delta, 0) - np.where(
            -w - delta > 0, -w - delta, 0
        )

        # Normalize the coefficients to unit length if nonzero
        coef /= np.linalg.norm(coef)

        # Return updated weights
        return coef

    else:
        # Raise an exception if no solution was found
        raise ValueError("No root was found for f")


def support_threshold(data, support, **kwargs):
    idx = np.argpartition(data.ravel(), data.shape[0] - support)
    data[idx[:-support]] = 0
    return data



import warnings

import numpy as np


def _process_parameter(parameter_name: str, parameter, default, n_views: int):
    if parameter is None:
        parameter = [default] * n_views
    elif not isinstance(parameter, (list, tuple)):
        parameter = [parameter] * n_views
    _check_parameter_number(parameter_name, parameter, n_views)
    return parameter


def _check_parameter_number(parameter_name: str, parameter, n_views: int):
    if len(parameter) != n_views:
        raise ValueError(
            f"number of views passed should match number of parameter {parameter_name}"
            f"len(views)={n_views} and "
            f"len({parameter_name})={len(parameter)}"
        )


def _check_Parikh2014(mus, lams, views):
    """Return index of the view which the data not matching the condition
    documented in Parikh 2014."""
    failed_check = [
        i
        for i, (mu, lam, view) in enumerate(zip(mus, lams, views))
        if mu < lam / np.linalg.norm(view) ** 2
    ]
    if failed_check:
        raise ValueError(
            "mu, lam, view not matching condition specified "
            "from Parikh 2014 (mu<lam/frobenius(views)**2)."
            "Index of view(s) not meeting the condition: "
            f"{failed_check}."
        )


def _check_batch_size(batch_size, latent_dimensions):
    """check batch size greater than number of latent dimensions and warn user otherwise"""
    if batch_size < latent_dimensions:
        warnings.warn(
            "Objective is unstable when batch size is less than the number of latent dimensions"
        )


def check_tsne_support(caller_name):
    try:
        import openTSNE
    except ImportError:
        raise ImportError(
            f"{caller_name} requires openTSNE. "
            "Please install openTSNE using `pip install openTSNE`"
        )


def check_umap_support(caller_name):
    try:
        import umap
    except ImportError:
        raise ImportError(
            f"{caller_name} requires umap. "
            "Please install umap using `pip install umap-learn`"
        )


def check_seaborn_support(caller_name):
    try:
        import seaborn
    except ImportError:
        raise ImportError(
            f"{caller_name} requires seaborn. "
            "Please install seaborn using `pip install seaborn`"
        )


def check_arviz_support(caller_name):
    try:
        import arviz as az
    except ImportError:
        raise ImportError(
            f"{caller_name} requires arviz. "
            "Please install arviz using `pip install arviz`"
        )



from typing import Iterable
import numpy as np
from sklearn import clone
from tqdm import tqdm


class DeflationMixin:
    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        views = self._validate_data(views)
        if isinstance(views, tuple):
            views = list(views)

        self.weights = [
            np.empty((view.shape[1], self.latent_dimensions)) for view in views
        ]

        for k in tqdm(
            range(self.latent_dimensions),
            desc="Latent Dimension",
            position=0,
            leave=True,
        ):
            component_weights = clone(self).set_params(latent_dimensions=1)._fit(views)

            for i, weight in enumerate(component_weights):
                self.weights[i][:, k] = weight.squeeze()

            if k < self.latent_dimensions - 1:
                pls = True  # self._get_tags().get("pls", False)
                views = deflate_views(views, component_weights, pls)

        return self


def deflate_views(
    residuals: Iterable[np.ndarray], weights: Iterable[np.ndarray], pls=False
):
    if pls:
        return [
            deflate_view_pls(residual, weight)
            for residual, weight in zip(residuals, weights)
        ]

    else:
        return [
            deflate_view_cca(residual, weight)
            for residual, weight in zip(residuals, weights)
        ]


def deflate_view_pls(residual: np.ndarray, weights: np.ndarray) -> np.ndarray:
    """
    Generalised deflation for PLS

    This method ensures orthogonal weights in the consecutive associative effects in each data modality

    """
    score = residual @ weights
    deflation_term = (score @ weights.T) / (weights.T @ weights)
    residual -= deflation_term
    return residual


def deflate_view_cca(residual: np.ndarray, weights: np.ndarray) -> np.ndarray:
    """
    PLS Mode-A deflation/ CCA deflation

    This method ensures orthogonal latent variables in the consecutive associative effects in each data modality.

    """
    score = residual @ weights
    residual -= score @ (score.T @ residual) / (score.T @ score)
    return residual



from typing import Union, Iterable

import numpy as np

from cca_zoo.linear._iterative._base import BaseIterative
from cca_zoo.linear._iterative._deflation import DeflationMixin
from cca_zoo.linear._search import _delta_search
from cca_zoo.linear._search import support_threshold
from cca_zoo.utils import _process_parameter


class SCCA_Span(DeflationMixin, BaseIterative):
    r"""
    Fits a Sparse CCA model using SpanCCA.

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} \|X_iw_i-X_jw_j\|^2 + \text{l1_ratio}\|w_i\|_1\}\\

        \text{subject to:}

        w_i^TX_i^TX_iw_i=1

    References
    ----------
    Asteris, Megasthenis, et al. "A simple and provable algorithm for sparse diagonal CCA." International Conference on Machine Learning. PMLR, 2016.
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        epochs: int = 100,
        copy_data=True,
        initialization: str = "pls",
        tol: float = 1e-3,
        regularisation="l0",
        tau: Union[Iterable[Union[float, int]], Union[float, int]] = None,
        rank=1,
        positive: Union[Iterable[bool], bool] = None,
        random_state=None,
        verbose=True,
        early_stopping=False,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            epochs=epochs,
            copy_data=copy_data,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            verbose=verbose,
            early_stopping=early_stopping,
        )
        self.tau = tau
        self.regularisation = regularisation
        self.rank = rank
        self.positive = positive

    def _check_params(self):
        """check number of views=2"""
        if self.n_views_ != 2:
            raise ValueError(f"SCCA_Span requires only 2 views")
        self.max_obj = 0
        if self.regularisation == "l0":
            self.update = support_threshold
        elif self.regularisation == "l1":
            self.update = _delta_search
        self.tau = _process_parameter("tau", self.tau, 0, self.n_views_)
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views_
        )

    def _update_weights(self, views: np.ndarray, i: int) -> None:
        """Update the weights for the i-th component.

        Args:
            views (np.ndarray): The input views as numpy arrays.
            i (int): The index of the component.
        """
        # if P, D, Q not initialised, initialise them
        if getattr(self, "P", None) is None:
            self._initialize_variables(views)
        if i == 0:
            # generate a random vector c
            c = np.random.randn(self.rank)
            c /= np.linalg.norm(c)
            # compute a = P D c
            a = self.P @ np.diag(self.D) @ c
            # apply the update function to a with tau[0]
            u = self.update(a, self.tau[0])
            u /= np.linalg.norm(u)
            # update the objective value and the weights if improved
            return u[:, np.newaxis]
        elif i == 1:
            b = self.Q @ np.diag(self.D) @ self.P.T @ self.weights[0]
            v = self.update(b, self.tau[1])
            v /= np.linalg.norm(v)
            return v

    def _initialize_variables(self, views):
        self.max_obj = [0, 0]
        cov = views[0].T @ views[1] / views[0].shape[0]
        # Perform SVD on im and obtain individual matrices
        P, D, Q = np.linalg.svd(cov, full_matrices=True)
        self.P = P[:, : self.rank]
        self.D = D[: self.rank]
        self.Q = Q[: self.rank, :].T



import warnings
from typing import Iterable

import numpy as np

from cca_zoo.linear._mcca import MCCA
from cca_zoo.utils import _process_parameter


class GRCCA(MCCA):
    """
    Grouped Regularized Canonical Correlation Analysis

    Parameters
    ----------
    latent_dimensions : int, default=1
        Number of latent dimensions to use
    copy_data : bool, default=True
        Whether to copy the data
    random_state : int, default=None
        Random state for initialisation
    eps : float, default=1e-3
        Tolerance for convergence
    c : float, default=0
        Regularization parameter for the group means
    mu : float, default=0
        Regularization parameter for the group sizes


    References
    ----------
    Tuzhilina, Elena, Leonardo Tozzi, and Trevor Hastie. "Canonical correlation analysis in high dimensions with structured regularization." Statistical Modelling (2021): 1471082X211041033.
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        eps=1e-3,
        c: float = 0,
        mu: float = 0,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            eps=eps,
            c=c,
            pca=False,
        )
        self.mu = mu

    def _check_params(self):
        self.mu = _process_parameter("mu", self.mu, 0, self.n_views_)
        self.c = _process_parameter("c", self.c, 0, self.n_views_)

    def fit(self, views: Iterable[np.ndarray], y=None, feature_groups=None, **kwargs):
        return super().fit(views, y=y, feature_groups=feature_groups, **kwargs)

    def _weights(self, eigvals, eigvecs, views, feature_groups=None, **kwargs):
        # Loop through c and add group means to splits if c > 0
        self.splits = [
            n_features + n_groups if c > 0 else n_features
            for n_features, n_groups, c in zip(self.n_features_, self.n_groups_, self.c)
        ]

        # Add zero at the beginning and compute cumulative sum of splits
        self.splits = np.insert(np.cumsum(self.splits), 0, 0)

        # Slice eigenvectors according to splits
        self.weights = [
            eigvecs[split:next_split]
            for split, next_split in zip(self.splits[:-1], self.splits[1:])
        ]

        # Adjust weights for each view based on group means and mu parameters
        for i, view in enumerate(views):
            if self.c[i] > 0:
                weights_1 = self.weights[i][: -self.n_groups_[i]]
                weights_2 = self.weights[i][-self.n_groups_[i] :]
                ids, unique_inverse, unique_counts, group_means = self._group_mean(
                    weights_1.T, feature_groups[i]
                )
                weights_1 = (weights_1 - group_means[:, unique_inverse].T) / self.c[i]
                mu = 1 if self.mu[i] == 0 else self.mu[i]
                weights_2 = weights_2 / np.sqrt(
                    mu * np.expand_dims(unique_counts, axis=1)
                )
                self.weights[i] = weights_1 + weights_2[feature_groups[i]]

    def _process_data(self, views, feature_groups=None, **kwargs):
        # Use all features if no feature groups are provided
        if feature_groups is None:
            warnings.warn(f"No feature groups provided, using all features")
            feature_groups = [np.ones(view.shape[1], dtype=int) for view in views]

        # Check that feature groups are integers
        for feature_group in feature_groups:
            assert np.issubdtype(
                feature_group.dtype, np.integer
            ), "feature groups must be integers"

        # Number of unique groups in each view
        self.n_groups_ = [np.unique(group).shape[0] for group in feature_groups]
        # Process each view and return a list of processed views and indices
        return [
            self._process_view(view, group, mu, c)
            for view, group, mu, c in zip(views, feature_groups, self.mu, self.c)
        ]

    def _process_view(self, view, group, mu, c):
        """
        Process a single view by subtracting group means and adding them as new features.

        Parameters
        ----------
        view : numpy array or array like with shape (n_samples, n_features)
            The view to be processed.

        group : numpy array or array like with shape (n_features,)
            The feature group labels for the view.

        mu : float
            The regularization parameter for the group means.

        c : float
            The regularization parameter for the view features.

        Returns
        -------
        view : numpy array with shape (n_samples, n_features + n_groups)
            The processed view with group means added as new features.
        """
        if c > 0:
            ids, unique_inverse, unique_counts, group_means = self._group_mean(
                view, group
            )
            mu = 1 if mu == 0 else mu
            view_1 = (view - group_means[:, unique_inverse]) / c
            view_2 = group_means / np.sqrt(mu / unique_counts)
            return np.hstack((view_1, view_2))
        else:
            return view

    def _more_tags(self):
        return {"multiview": True}

    @staticmethod
    def _group_mean(view, group):
        """
        Compute the mean of each feature group in a view.

        Parameters
        ----------
        view : numpy array or array like with shape (n_samples, n_features)
            The view to compute the group means from.

        group : numpy array or array like with shape (n_features,)
            The feature group labels for the view.

        Returns
        -------
        ids : numpy array with shape (n_groups,)
            The unique feature group ids.

        unique_inverse : numpy array with shape (n_features,)
            The indices to reconstruct the original group array from the unique ids.

        unique_counts : numpy array with shape (n_groups,)
            The number of occurrences of each unique id in the group array.

        group_means : numpy array with shape (n_samples, n_groups)
            The mean of each feature group in the view.
        """
        ids, unique_inverse, unique_counts = np.unique(
            group, return_inverse=True, return_counts=True
        )
        # Use axis argument to compute mean along columns for each group
        group_means = np.array([view[:, group == id].mean(axis=1) for id in ids]).T
        return ids, unique_inverse, unique_counts, group_means



from .explained_variance import ExplainedVarianceDisplay
from .scores import ScoreDisplay
from .explained_covariance import ExplainedCovarianceDisplay
from .weights import WeightHeatmapDisplay
from .correlation import CorrelationHeatmapDisplay
from .covariance import CovarianceHeatmapDisplay
from .tsne_scores import TSNEScoreDisplay
from .umap_scores import UMAPScoreDisplay
from .inference import WeightInferenceDisplay

__all__ = [
    "ExplainedVarianceDisplay",
    "ScoreDisplay",
    "ExplainedCovarianceDisplay",
    "WeightHeatmapDisplay",
    "CorrelationHeatmapDisplay",
    "CovarianceHeatmapDisplay",
    "TSNEScoreDisplay",
    "UMAPScoreDisplay",
    "WeightInferenceDisplay",
]



import matplotlib.pyplot as plt
import seaborn as sns

from cca_zoo.utils.check_values import check_seaborn_support


class ScoreDisplay:
    """
    Display the scores of a model
    """

    def __init__(
        self,
        train_scores,
        test_scores,
        labels=None,
        test_labels=None,
        separate=False,
        **kwargs,
    ):
        self.train_scores = train_scores
        self.test_scores = test_scores
        self.train_labels = labels
        self.test_labels = test_labels
        self.kwargs = kwargs
        self.separate = separate

    def _validate_plot_params(self):
        check_seaborn_support("CorrelationHeatmapDisplay")

    @classmethod
    def from_estimator(cls, model, train_views, test_views=None, **kwargs):
        train_scores = model.transform(train_views)
        if test_views is not None:
            test_scores = model.transform(test_views)
        else:
            test_scores = None
        return cls.from_scores(train_scores, test_scores, **kwargs)

    @classmethod
    def from_scores(cls, train_scores, test_scores=None, **kwargs):
        return cls(train_scores, test_scores, **kwargs)

    def plot(self):
        dimensions = self.train_scores[0].shape[1]

        if self.separate:
            fig, axarr = plt.subplots(
                dimensions, 2
            )  # Create two columns for each dimension

            for i in range(dimensions):
                # Plotting training scores
                sns.scatterplot(
                    x=self.train_scores[0][:, i],
                    y=self.train_scores[1][:, i],
                    hue=self.train_labels,
                    ax=axarr[i, 0],
                    alpha=1.0,
                    label="Train",
                    **self.kwargs,
                )

                axarr[i, 0].set_title(f"Train - Latent Dimension {i+1}")

                # Plotting testing scores if available
                if self.test_scores is not None:
                    sns.scatterplot(
                        x=self.test_scores[0][:, i],
                        y=self.test_scores[1][:, i],
                        hue=self.test_labels,
                        ax=axarr[i, 1],
                        label="Test",
                        **self.kwargs,
                    )

                    axarr[i, 1].set_title(f"Test - Latent Dimension {i+1}")
        else:
            fig, ax = plt.subplots(dimensions)

            for i in range(dimensions):
                sns.scatterplot(
                    x=self.train_scores[0][:, i],
                    y=self.train_scores[1][:, i],
                    hue=self.train_labels,
                    ax=ax[i],
                    alpha=0.1 if self.test_scores is not None else 1.0,
                    label="Train" if self.test_scores is not None else None,
                    **self.kwargs,
                )

                ax[i].set_title(f"Latent Dimension {i+1}")

                if self.test_scores is not None:
                    sns.scatterplot(
                        x=self.test_scores[0][:, i],
                        y=self.test_scores[1][:, i],
                        hue=self.test_labels,
                        ax=ax[i],
                        label="Test",
                        **self.kwargs,
                    )
                    sns.move_legend(ax[i], "lower center", ncol=2)

        plt.tight_layout()
        self.figure_ = fig
        return self



import torch

from ._dcca import DCCA


class BarlowTwins(DCCA):
    """
    A class used to fit a Barlow Twins model.

    Barlow Twins is a self-supervised learning method that applies redundancy-reduction
    to learn representations that are invariant to distortions of the input sample.

    References
    ----------
    Zbontar, Jure, et al. "Barlow twins: Self-supervised learning via redundancy reduction." arXiv preprint arXiv:2103.03230 (2021).

    """

    def __init__(
        self,
        latent_dimensions: int,
        encoders=None,
        lam=1,
        **kwargs,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions, encoders=encoders, **kwargs
        )
        self.lam = lam  # the lambda parameter for the off-diagonal terms of the cross-covariance matrix
        self.bns = torch.nn.ModuleList(
            [
                torch.nn.BatchNorm1d(latent_dimensions, affine=False)
                for _ in self.encoders
            ]
        )  # a list of batch normalization layers for each encoder

    def forward(self, views, **kwargs):
        z = []
        for i, (encoder, bn) in enumerate(zip(self.encoders, self.bns)):
            z.append(bn(encoder(views[i])))  # encode and normalize each view
        return z  # return a list of normalized latent representations

    def loss(self, batch, **kwargs):
        z = self(batch["views"])  # get the latent representations
        cross_cov = (
            z[0].T @ z[1] / z[0].shape[0]
        )  # compute the cross-covariance matrix between the two views
        invariance = torch.sum(
            torch.pow(1 - torch.diag(cross_cov), 2)
        )  # compute the invariance term as the sum of squared differences from 1 on the diagonal
        covariance = torch.sum(
            torch.triu(torch.pow(cross_cov, 2), diagonal=1)
        ) + torch.sum(
            torch.tril(torch.pow(cross_cov, 2), diagonal=-1)
        )  # compute the covariance term as the sum of squared values on the off-diagonal
        return {
            "objective": invariance
            + self.lam
            * covariance,  # return the objective value as a combination of invariance and covariance terms
            "invariance": invariance,  # return the invariance term
            "covariance": covariance,  # return the covariance term
        }



import torch
import torch.nn.functional as F

from ._dcca_noi import DCCA_NOI


class DCCA_SDL(DCCA_NOI):
    """
    A class used to fit a Deep CCA by Stochastic Decorrelation model.

    References
    ----------
    Chang, Xiaobin, Tao Xiang, and Timothy M. Hospedales. "Scalable and effective deep CCA via soft decorrelation." Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition. 2018.

    """

    def __init__(
        self,
        latent_dimensions: int,
        encoders=None,
        r: float = 0,
        rho: float = 0.2,
        eps: float = 1e-5,
        shared_target: bool = False,
        lam=0.5,
        **kwargs
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            encoders=encoders,
            r=r,
            rho=rho,
            eps=eps,
            shared_target=shared_target,
            **kwargs
        )
        self.c = None
        self.cross_cov = None
        self.lam = lam
        self.bns = torch.nn.ModuleList(
            [
                torch.nn.BatchNorm1d(latent_dimensions, affine=False)
                for _ in self.encoders
            ]
        )

    def forward(self, views, **kwargs):
        z = []
        for i, (encoder, bn) in enumerate(zip(self.encoders, self.bns)):
            z.append(bn(encoder(views[i])))
        return z

    def loss(self, batch, **kwargs):
        z = self(batch["views"])
        l2_loss = F.mse_loss(z[0], z[1])
        SDL_loss = self._sdl_loss(self.covs)
        loss = l2_loss + self.lam * SDL_loss
        return {"objective": loss, "l2": l2_loss, "sdl": SDL_loss}

    def _sdl_loss(self, covs):
        loss = 0
        for cov in covs:
            sgn = torch.sign(cov)
            sgn.fill_diagonal_(0)
            loss += torch.mean(cov * sgn)
        return loss



from typing import Iterable, Union

import numpy as np
from sklearn.metrics import pairwise_kernels
from sklearn.neighbors import NearestNeighbors

from cca_zoo._base import BaseModel
from cca_zoo.utils.check_values import _process_parameter


class NCCA(BaseModel):
    """
    A class used to fit nonparametric (NCCA) model. This model extends CCA to nonlinear relationships by using local linear projections based on nearest neighbors.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    accept_sparse : bool, optional
        Whether to accept sparse data as input, by default False
    random_state : Union[int, np.random.RandomState], optional
        Random seed for reproducibility, by default None
    nearest_neighbors : int, optional
        Number of nearest neighbors to use for local linear projections, by default None. If None, it will use the square root of the number of samples.
    gamma : Iterable[float], optional
        Bandwidth parameter or list of parameters for the RBF kernel for each view, by default None. If None, it will use the median heuristic.


    References
    ----------
    Michaeli, Tomer, Weiran Wang, and Karen Livescu. "Nonparametric canonical correlation analysis." International conference on machine learning. PMLR, 2016.

    Example
    -------
    >>> from cca_zoo.linear import NCCA
    >>> X1 = np.random.rand(10,5)
    >>> X2 = np.random.rand(10,5)
    >>> model = NCCA()
    >>> model.fit((X1,X2)).score((X1,X2))
    array([1.])
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        accept_sparse=False,
        random_state: Union[int, np.random.RandomState] = None,
        nearest_neighbors=None,
        gamma: Iterable[float] = None,
    ):
        # Call the parent class constructor
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
        )
        # Store the nearest neighbors and gamma parameters
        self.nearest_neighbors = nearest_neighbors
        self.gamma = gamma

    def _check_params(self):
        # Process the nearest neighbors and gamma parameters for each view
        self.nearest_neighbors = _process_parameter(
            "nearest_neighbors", self.nearest_neighbors, 1, self.n_views_
        )
        self.gamma = _process_parameter("gamma", self.gamma, None, self.n_views_)
        # Use RBF kernel as default for each view
        self.kernel = _process_parameter("kernel", None, "rbf", self.n_views_)

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        # Validate the input data
        views = self._validate_data(views)
        # Check the parameters
        self._check_params()
        # Store the training views
        self.train_views = views
        # Fit a nearest neighbors model for each view
        self.knns = [
            NearestNeighbors(n_neighbors=self.nearest_neighbors[i]).fit(view)
            for i, view in enumerate(views)
        ]
        # Find the nearest neighbors for each view
        NNs = [
            self.knns[i].kneighbors(view, self.nearest_neighbors[i])
            for i, view in enumerate(views)
        ]
        # Compute the kernel matrices for each view
        kernels = [self._get_kernel(i, view) for i, view in enumerate(self.train_views)]
        # Fill the weight matrices with the kernel values
        self.Ws = [fill_w(kernel, inds) for kernel, (dists, inds) in zip(kernels, NNs)]
        # Normalize the weight matrices by row and column sums
        self.Ws = [
            self.Ws[0] / self.Ws[0].sum(axis=1, keepdims=True),
            self.Ws[1] / self.Ws[1].sum(axis=0, keepdims=True),
        ]
        # Compute the cross-covariance matrix between the weight matrices
        S = self.Ws[0] @ self.Ws[1]
        # Perform singular value decomposition on the cross-covariance matrix
        U, S, Vt = np.linalg.svd(S)
        # Compute the canonical score vectors for each view
        self.f = U[:, 1 : self.latent_dimensions + 1] * np.sqrt(self.n_samples_)
        self.g = Vt[1 : self.latent_dimensions + 1, :].T * np.sqrt(self.n_samples_)
        # Store the canonical correlations
        self.S = S[1 : self.latent_dimensions + 1]
        return self

    def transform(self, views: Iterable[np.ndarray], **kwargs):
        # Find the nearest neighbors for each view
        nns = [
            self.knns[i].kneighbors(view, self.nearest_neighbors[i])
            for i, view in enumerate(views)
        ]
        # Compute the kernel matrices between the training and test views
        kernels = [
            self._get_kernel(i, self.train_views[i], Y=view)
            for i, view in enumerate(views)
        ]
        # Fill the weight matrices with the kernel values
        Wst = [fill_w(kernel, inds) for kernel, (dists, inds) in zip(kernels, nns)]
        # Normalize the weight matrices by row sums
        Wst = [
            Wst[0] / Wst[0].sum(axis=1, keepdims=True),
            Wst[1] / Wst[1].sum(axis=1, keepdims=True),
        ]
        # Compute the cross-covariance matrix between the weight matrices and the training weight matrices
        St = [Wst[0] @ self.Ws[1], Wst[1] @ self.Ws[0]]
        # Project the cross-covariance matrix onto the canonical score vectors and normalize by the canonical correlations
        return St[0] @ self.g * (1 / self.S), St[1] @ self.f * (1 / self.S)

    def _get_kernel(self, view, X, Y=None):
        # Get the gamma parameter for the RBF kernel
        params = {
            "gamma": self.gamma[view],
        }
        # Compute the pairwise kernel values between views and Y using the specified kernel function and parameters
        return pairwise_kernels(
            X, Y, metric=self.kernel[view], filter_params=True, **params
        )


def fill_w(kernels, inds):
    # Create an empty matrix with the same shape as kernels
    w = np.zeros_like(kernels)
    # For each row, fill the corresponding columns with the kernel values
    for i, ind in enumerate(inds):
        w[ind, i] = kernels[ind, i]
        # Transpose the matrix to match the original orientation
    return w.T



from ._ey import CCA_EY, PLS_EY
from ._gha import CCA_GHA

from ._stochasticpls import PLSStochasticPower
from ._svd import CCA_SVD, PLS_SVD

__all__ = [
    "CCA_EY",
    "PLS_EY",
    "CCA_GHA",
    "CCA_SVD",
    "PLS_SVD",
    "PLSStochasticPower",
]



from typing import Iterable

import jax.numpy as jnp
import numpy as np
import numpyro
import numpyro.distributions as dist
from jax.random import PRNGKey
from numpyro.infer import MCMC, NUTS, Predictive
from sklearn.utils.validation import check_is_fitted

from cca_zoo._base import BaseModel


class ProbabilisticCCA(BaseModel):
    """
    A class used to fit a Probabilistic CCA model using variational inference.

    Probabilistic CCA is a generative model that assumes each view of data is generated from a shared latent variable z and some view-specific parameters (mu: mean, psi: covariance, W: weight matrix). The model can be written as:

    z ~ N(0, I)
    x_i ~ N(W_i z + mu_i, psi_i)

    The model parameters and the latent variables are inferred using MCMC sampling with the NUTS algorithm.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state, by default 0
    num_samples : int, optional
        Number of samples to use in MCMC, by default 100
    num_warmup : int, optional
        Number of warmup samples to use in MCMC, by default 100


    References
    ----------
    Bach, Francis R., and Michael I. Jordan. "A probabilistic interpretation of canonical correlation analysis." (2005).
    Wang, Chong. "Variational Bayesian approach to canonical correlation analysis." IEEE Transactions on Neural Networks 18.3 (2007): 905-910.

    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state: int = 0,
        num_samples=100,
        num_warmup=100,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            accept_sparse=False,
            random_state=random_state,
        )
        self.num_samples = num_samples
        self.num_warmup = num_warmup
        self.rng_key = PRNGKey(random_state)

    def fit(self, views: Iterable[np.ndarray], y=None):
        """
        Infer the parameters and latent variables of the Probabilistic CCA model.

        Parameters
        ----------
        views : Iterable[np.ndarray]
            A list or tuple of numpy arrays or array likes with the same number of rows (samples)

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        views = self._validate_data(views)
        # Initialize a NUTS sampler with the model function
        nuts_kernel = NUTS(self._model)
        # Run MCMC sampling with the specified number of samples and warmup steps
        self.mcmc = MCMC(
            nuts_kernel, num_samples=self.num_samples, num_warmup=self.num_warmup
        )
        # Run the sampler on the data and store the posterior samples
        self.mcmc.run(self.rng_key, views)
        self.posterior_samples = self.mcmc.get_samples()
        return self

    def transform(self, views: Iterable[np.ndarray], y=None):
        """
        Predict the latent variables that generate the data in views using the sampled model parameters.

        Parameters
        ----------
        views : Iterable[np.ndarray]
            A list or tuple of numpy arrays or array likes with the same number of rows (samples)

        Returns
        -------
        z : np.ndarray
            An array of shape (n_samples, latent_dimensions) containing the predicted latent variables for each sample.
        """
        # Check if the model has been fitted
        check_is_fitted(self, attributes=["posterior_samples"])
        # Use the predictive function to generate samples of z from the posterior distribution
        return Predictive(self._model, self.posterior_samples, return_sites=["z"])(
            self.rng_key, views
        )["z"]

    def _model(self, views: Iterable[np.ndarray]):
        n = views[0].shape[0]
        p = [view.shape[1] for view in views]
        # parameter representing the mean of column in each view of data
        mu = [
            numpyro.sample(
                "mu_" + str(i), dist.MultivariateNormal(0.0, 10 * jnp.eye(p_))
            )
            for i, p_ in enumerate(p)
        ]
        # parameter representing the within view variance for each view of data
        psi = [
            numpyro.sample("psi_" + str(i), dist.LKJCholesky(p_))
            for i, p_ in enumerate(p)
        ]
        # parameter representing weights applied to latent variables
        with numpyro.plate("plate_views", self.latent_dimensions):
            self.weights_list = [
                numpyro.sample(
                    "W_" + str(i),
                    dist.MultivariateNormal(0.0, 10 * jnp.diag(jnp.ones(p_))),
                )
                for i, p_ in enumerate(p)
            ]
        with numpyro.plate("plate_i", n):
            # sample from latent z: the latent variables of the model
            z = numpyro.sample(
                "z",
                dist.MultivariateNormal(
                    0.0, jnp.diag(jnp.ones(self.latent_dimensions))
                ),
            )
            # sample from multivariate normal and observe data
            [
                numpyro.sample(
                    "obs" + str(i),
                    dist.MultivariateNormal((z @ W_) + mu_, scale_tril=psi_),
                    obs=X_,
                )
                for i, (X_, psi_, mu_, W_) in enumerate(
                    zip(views, psi, mu, self.weights_list)
                )
            ]

    def _more_tags(self):
        return {"probabilistic": True}
